from ptls.data_load.iterable_processing.to_torch_tensor import ToTorch
import numpy as np
import torch


def get_data():
    return [
        {
            'client_id': 1,
            'trans_date': np.array([0., 1., 2., 5., 8., 9., 11., 15.]),
            'small_group': np.array([0, 1, 2, 5, 8, 9, 11, 15])
        }
    ]


def test_to_torch_filter():
    data_before = get_data()
    data_after = list(ToTorch()(data_before))[0]
    data_before = data_before[0]
    for k, v in data_after.items():
        casted = torch.from_numpy(data_before[k]) if isinstance(data_before[k], np.ndarray) else data_before[k]
        if isinstance(v, torch.Tensor):
            assert torch.equal(v, casted)
        else:
            assert v == casted



"""Test examples from docs/frames/common_usage.md
"""
from functools import partial

import pytorch_lightning as pl
import torch
from sklearn.model_selection import train_test_split

from ptls.data_load.datasets.dataloaders import inference_data_loader
from ptls.frames.coles import CoLESModule
from ptls.frames.coles import ColesDataset
from ptls.frames.coles.split_strategy import SampleSlices
from ptls.frames.supervised import SequenceToTarget
from ptls.nn import TrxEncoder, RnnSeqEncoder
from ptls.frames import PtlsDataModule


def test_dataset_way():
    # Makes 1000 samples with `mcc_code` and `amount` features and seq_len randomly sampled in range (100, 200)
    dataset = [{
        'mcc_code': torch.randint(1, 10, (seq_len,)),
        'amount': torch.randn(seq_len),
        'event_time': torch.arange(seq_len),
    } for seq_len in torch.randint(100, 200, (1000,))]

    # split 10% for validation
    train_data, valid_data = train_test_split(dataset, test_size=0.1)

    # datasets
    splitter=SampleSlices(split_count=5, cnt_min=10, cnt_max=20)
    train_dataset = ColesDataset(data=train_data, splitter=splitter)
    valid_dataset = ColesDataset(data=valid_data, splitter=splitter)

    train_dataloader = torch.utils.data.DataLoader(
        dataset=train_dataset,
        collate_fn=train_dataset.collate_fn,
        shuffle=True,
        num_workers=4,
        batch_size=32,
    )
    valid_dataloader = torch.utils.data.DataLoader(
        dataset=valid_dataset,
        collate_fn=valid_dataset.collate_fn,
        shuffle=False,
        num_workers=0,
        batch_size=32,
    )

    seq_encoder = RnnSeqEncoder(
        trx_encoder=TrxEncoder(
            embeddings={'mcc_code': {'in': 10, 'out': 4}},
            numeric_values={'amount': 'identity'},
        ),
        hidden_size=16,  # this is final embedding size
    )

    coles_module = CoLESModule(
        seq_encoder=seq_encoder,
        optimizer_partial=partial(torch.optim.Adam, lr=0.001),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=1, gamma=0.9),
    )

    trainer = pl.Trainer(max_epochs=3)
    trainer.fit(coles_module, train_dataloader, valid_dataloader)

    # inference
    inference_dataloader = inference_data_loader(dataset, num_workers=4, batch_size=256)
    model = SequenceToTarget(seq_encoder)
    trainer = pl.Trainer()
    embeddings = torch.vstack(trainer.predict(model, inference_dataloader))
    assert embeddings.size() == (1000, 16)


def test_datamodule_way():
    # Makes 1000 samples with `mcc_code` and `amount` features and seq_len randomly sampled in range (100, 200)
    dataset = [{
        'mcc_code': torch.randint(1, 10, (seq_len,)),
        'amount': torch.randn(seq_len),
        'event_time': torch.arange(seq_len),
    } for seq_len in torch.randint(100, 200, (1000,))]

    # split 10% for validation
    train_data, valid_data = train_test_split(dataset, test_size=0.1)

    # datasets
    splitter=SampleSlices(split_count=5, cnt_min=10, cnt_max=20)
    train_dataset = ColesDataset(data=train_data, splitter=splitter)
    valid_dataset = ColesDataset(data=valid_data, splitter=splitter)

    datamodule = PtlsDataModule(
        train_data=train_dataset,
        train_batch_size=32,
        train_num_workers=4,
        valid_data=valid_dataset,
        valid_num_workers=0,
    )

    seq_encoder = RnnSeqEncoder(
        trx_encoder=TrxEncoder(
            embeddings={'mcc_code': {'in': 10, 'out': 4}},
            numeric_values={'amount': 'identity'},
        ),
        hidden_size=16,  # this is final embedding size
    )

    coles_module = CoLESModule(
        seq_encoder=seq_encoder,
        optimizer_partial=partial(torch.optim.Adam, lr=0.001),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=1, gamma=0.9),
    )

    trainer = pl.Trainer(max_epochs=3)
    trainer.fit(coles_module, datamodule)

    # inference
    inference_dataloader = inference_data_loader(dataset, num_workers=4, batch_size=256)
    model = SequenceToTarget(seq_encoder)
    trainer = pl.Trainer()
    embeddings = torch.vstack(trainer.predict(model, inference_dataloader))
    assert embeddings.size() == (1000, 16)



from typing import List

import torch
import pandas as pd

from ptls.preprocessing.base import ColTransformer
from ptls.preprocessing.pandas.col_transformer import ColTransformerPandasMixin


class UserGroupTransformer(ColTransformerPandasMixin, ColTransformer):
    """Groups transactions by user. Splits it by features.

    'event_time' column should be in dataset. We use it to order transactions

    Parameters
    ----------
    col_name_original:
        Column name with user_id - key for grouping
    cols_first_item:
        Only first value will be taken for these columns.
        All values as tensor will be taken for other columns
    return_records:
        False: Result is a dataframe. Use `.to_dict(orient='records')` to transform it to `ptls` format.
        True: Result is a list of dicts - `ptls` format
    """
    def __init__(self,
                 col_name_original: str,
                 cols_first_item: List[str] = None,
                 return_records: bool = False,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target=None,
            is_drop_original_col=False,
        )
        self.cols_first_item = cols_first_item if cols_first_item is not None else []
        self.return_records = return_records

    def fit(self, x):
        # super().fit(x)
        if self.col_name_original not in x.columns:
            raise AttributeError(f'col_name_original="{self.col_name_original}" not in source dataframe. '
                                 f'Found {x.columns}')
        if 'event_time' not in x.columns:
            raise AttributeError(f'"event_time" not in source dataframe. '
                                 f'Found {x.columns}')
        return self

    def df_to_feature_arrays(self, df):
        return pd.Series({k: torch.from_numpy(v.values) if k not in self.cols_first_item else v.iloc[0]
                          for k, v in df.to_dict(orient='series').items()})

    def transform(self, x: pd.DataFrame):
        x = self.attach_column(x, x['event_time'].rename('et_index')).set_index([self.col_name_original, 'et_index'])
        x = x.sort_index().groupby(self.col_name_original)
        x = x.apply(self.df_to_feature_arrays).reset_index()

        # we don't heed to drop original column in this transformer
        # x = super().transform(x)
        if self.return_records:
            x = x.to_dict(orient='records')
        return x



import torch
from torch.nn import functional as F


class MarginLoss(torch.nn.Module):

    """
    Margin loss

    "Sampling Matters in Deep Embedding Learning", ICCV 2017
    https://arxiv.org/abs/1706.07567

    """

    def __init__(self, pair_selector, margin=1, beta=1.2):
        super(MarginLoss, self).__init__()
        self.margin = margin
        self.beta = beta
        self.pair_selector = pair_selector

    def forward(self, embeddings, target):

        positive_pairs, negative_pairs = self.pair_selector.get_pairs(embeddings, target)

        d_ap = F.pairwise_distance(embeddings[positive_pairs[:, 0]], embeddings[positive_pairs[:, 1]])
        d_an = F.pairwise_distance(embeddings[negative_pairs[:, 0]], embeddings[negative_pairs[:, 1]])

        pos_loss = torch.clamp(d_ap - self.beta + self.margin, min=0.0)
        neg_loss = torch.clamp(self.beta - d_an + self.margin, min=0.0)

        loss = torch.cat([pos_loss, neg_loss], dim=0)

        return loss.sum()


"""
This class inherited from ColesDataModuleTrain. Just rewrited collate function
"""
import warnings

"""
Creates data loaders for contrastive learning
Each record pass throw the sampler and splits into views of the main record
X - is feature_arrays, sub-samples of main event sequences according
    split strategy: /dltranz/metric_learn/dataset/split_strategy.py

Input:            -> Output
dict of arrays    ->  X,
client_1          -> [client_1_smpl_1, client_1_smpl_2, ..., client_1_smpl_n]
client_2          -> [client_2_smpl_1, client_2_smpl_2, ..., client_2_smpl_n]
...
client_k          -> [client_k_smpl_1, client_k_smpl_2, ..., client_k_smpl_n]
"""


import logging

import torch

from collections import defaultdict
from torch.utils.data import DataLoader

from ptls.data_load.padded_batch import PaddedBatch
from ptls.data_load.data_module.coles_data_module import ColesDataModuleTrain

import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')

logger = logging.getLogger(__name__)


def collate_fn(batch):
    """
    batch = [
        [PaddedBatch, PaddedBatch, ..., PaddedBatch],
        [PaddedBatch, PaddedBatch, ..., PaddedBatch],
        ...
        [PaddedBatch, PaddedBatch, ..., PaddedBatch]
    ]
    """
    split_count = len(batch[0])

    sequences = [defaultdict(list) for _ in range(split_count)]
    lengths = torch.zeros((len(batch), split_count), dtype=torch.int)

    for i, client_data in enumerate(batch):
        for j, subseq in enumerate(client_data):
            for k, v in subseq.items():
                sequences[j][k].append(v)
                lengths[i][j] = v.shape[0]

    # adding padds in each split
    padded_batch = []
    for j, seq in enumerate(sequences):
        for k, v in seq.items():
            seq[k] = torch.nn.utils.rnn.pad_sequence(v, batch_first=True)
        padded_batch.append(PaddedBatch(seq, lengths[:, j].long()))

    return tuple(padded_batch)


class CpcV2DataModuleTrain(ColesDataModuleTrain):
    def __init__(self, type, setup, train, valid, pl_module):
        warnings.warn('Use `ptls.frames.PtlsDataModule` '
                      'with `ptls.frames.cpc.CpcV2Dataset` or `ptls.frames.cpc.CpcV2IterableDataset`',
                      DeprecationWarning)
        super().__init__(type, setup, train, valid, pl_module)

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=collate_fn,
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf.num_workers,
            batch_size=self.train_conf.batch_size,
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.valid_dataset,
            collate_fn=collate_fn,
            num_workers=self.valid_conf.num_workers,
            batch_size=self.valid_conf.batch_size,
        )



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl

from ptls.data_load.padded_batch import PaddedBatch

class DropoutEncoder(nn.Module):
    def __init__(self, p):
        super().__init__()
        self.p = p

    def forward(self, x):
        if self.training:
            mask = torch.FloatTensor(x.shape[1]).uniform_(0, 1) <= self.p
            x = x.masked_fill(mask.to(x.device), 0)
        return x


class Squeeze(nn.Module):
    """Use torch.nn.Flatten(start_dim=) instead

    """
    def forward(self, x: torch.Tensor):
        return x.squeeze()


class CatLayer(nn.Module):
    def __init__(self, left_tail, right_tail):
        super().__init__()
        self.left_tail = left_tail
        self.right_tail = right_tail

    def forward(self, x):
        l, r = x
        t = torch.cat([self.left_tail(l), self.right_tail(r)], axis=1)
        return t


class MLP(nn.Module):
    def __init__(self, input_size, params):
        super().__init__()
        self.input_size = input_size
        self.use_batch_norm = params.get('use_batch_norm', True)

        layers = []
        if self.use_batch_norm:
            layers.append(nn.BatchNorm1d(input_size))
        layers_size = [input_size] + list(params.hidden_layers_size)
        for size_in, size_out in zip(layers_size[:-1], layers_size[1:]):
            layers.append(nn.Linear(size_in, size_out))
            layers.append(nn.ReLU())
            if params.drop_p:
                layers.append(nn.Dropout(params.drop_p))
            if self.use_batch_norm:
                layers.append(nn.BatchNorm1d(size_out))
            self.output_size = layers_size[-1]

        if params.get('objective', None) == 'classification':
            head_output_size = params.get('num_classes', 1)
            if head_output_size == 1:
                h = nn.Sequential(nn.Linear(layers_size[-1], head_output_size), nn.Sigmoid(), Squeeze())
            else:
                h = nn.Sequential(nn.Linear(layers_size[-1], head_output_size), nn.LogSoftmax(dim=1))
            layers.append(h)
            self.output_size = head_output_size

        elif params.get('objective', None) == 'multilabel_classification':
            head_output_size = params.num_classes
            h = nn.Sequential(nn.Linear(layers_size[-1], head_output_size), nn.Sigmoid())
            layers.append(h)
            self.output_size = head_output_size

        elif params.get('objective', None) == 'regression':
            h = nn.Sequential(nn.Linear(layers_size[-1], 1), Squeeze())
            layers.append(h)
            self.output_size = 1

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class TabularRowEncoder(torch.nn.Module):
    def __init__(self, input_dim, cat_dims, cat_idxs, cat_emb_dim):
        """ This is an embedding module for an entier set of features

        Parameters
        ----------
        input_dim : int
            Number of features coming as input (number of columns)
        cat_dims : list of int
            Number of modalities for each categorial features
            If the list is empty, no embeddings will be done
        cat_idxs : list of int
            Positional index for each categorical features in inputs
        cat_emb_dim : int or list of int
            Embedding dimension for each categorical features
            If int, the same embdeding dimension will be used for all categorical features
        """
        super().__init__()
        if cat_dims == [] or cat_idxs == []:
            self.skip_embedding = True
            self.post_embed_dim = input_dim
            return

        self.skip_embedding = False
        self.cat_idxs = cat_idxs
        if isinstance(cat_emb_dim, int):
            self.cat_emb_dims = [cat_emb_dim] * len(cat_idxs)
        else:
            self.cat_emb_dims = cat_emb_dim

        # check that all embeddings are provided
        if len(self.cat_emb_dims) != len(cat_dims):
            msg = """ cat_emb_dim and cat_dims must be lists of same length, got {len(self.cat_emb_dims)}
                      and {len(cat_dims)}"""
            raise ValueError(msg)
        self.post_embed_dim = int(input_dim + np.sum(self.cat_emb_dims) - len(self.cat_emb_dims))

        self.embeddings = torch.nn.ModuleList()

        # Sort dims by cat_idx
        sorted_idxs = np.argsort(cat_idxs)
        cat_dims = [cat_dims[i] for i in sorted_idxs]
        self.cat_emb_dims = [self.cat_emb_dims[i] for i in sorted_idxs]

        for cat_dim, emb_dim in zip(cat_dims, self.cat_emb_dims):
            self.embeddings.append(torch.nn.Embedding(cat_dim, emb_dim))
            # TODO: NoisyEmbeddings

    def forward(self, x):
        """
        Apply embdeddings to inputs
        Inputs should be (batch_size, input_dim)
        Outputs will be of size (batch_size, self.post_embed_dim)
        """
        if self.skip_embedding:
            # no embeddings required
            return x

        cols = []
        prev_cat_idx = -1
        for cat_feat_counter, cat_idx in enumerate(self.cat_idxs):
            if cat_idx > prev_cat_idx + 1:
                cols.append(x[:, prev_cat_idx + 1:cat_idx].float())
            cols.append(self.embeddings[cat_feat_counter](x[:, cat_idx].long()))
            prev_cat_idx = cat_idx

        if prev_cat_idx + 1 < x.shape[1]:
            cols.append(x[:, prev_cat_idx + 1:].float())

        # concat
        post_embeddings = torch.cat(cols, dim=1)
        return post_embeddings

    @property
    def output_size(self):
        return self.post_embed_dim


class DistributionTargetHead(torch.nn.Module):
    def __init__(self, in_size=256, num_distr_classes_pos=4, num_distr_classes_neg=14, pos=True, neg=True):
        super().__init__()
        self.pos, self.neg = pos, neg
        self.dense1 = torch.nn.Linear(in_size, 128)

        self.dense2_neg = torch.nn.Linear(128, num_distr_classes_neg) if self.neg else None
        self.dense2_pos = torch.nn.Linear(128, num_distr_classes_pos) if self.pos else None

        self.sigmoid = torch.nn.Sigmoid()
        self.relu = torch.nn.ReLU()

    def forward(self, x):
        out1 = self.relu(self.dense1(x))
        out2_pos = out2_neg = 0
        if self.pos:
            out2_pos = self.dense2_pos(out1)
        if self.neg:
            out2_neg = self.dense2_neg(out1)
        return out2_neg, out2_pos


class RegressionTargetHead(torch.nn.Module):
    def __init__(self, in_size=256, gates=True, pos=True, neg=True, pass_samples=True):
        super().__init__()
        self.pos, self.neg = pos, neg
        self.pass_samples = pass_samples
        self.dense1 = torch.nn.Linear(in_size, 64)
        if pass_samples:
            self.dense2_neg = torch.nn.Linear(64, 15) if self.neg else None
            self.dense2_pos = torch.nn.Linear(64, 15) if self.pos else None
        else:
            self.dense2_neg = torch.nn.Linear(64, 16) if self.neg else None
            self.dense2_pos = torch.nn.Linear(64, 16) if self.pos else None
        self.dense3_neg = torch.nn.Linear(16, 1) if self.neg else None
        self.dense3_pos = torch.nn.Linear(16, 1) if self.pos else None
        self.sigmoid = torch.nn.Sigmoid()
        self.relu = torch.nn.ReLU()
        self.gates = gates

    def forward(self, x, neg_sum_logs=None, pos_sum_logs=None):
        out1 = self.relu(self.dense1(x))
        if self.gates:
            out1 = out1.detach()

        out3_neg = out3_pos = 0
        if self.pos:
            out2_pos = self.relu(self.dense2_pos(out1))
            if self.pass_samples:
                out2_pos = torch.cat((out2_pos, pos_sum_logs), dim=1).float()
            out3_pos = self.dense3_pos(out2_pos)
            out3_pos = self.sigmoid(out3_pos) if self.gates else out3_pos
        if self.neg:
            out2_neg = self.relu(self.dense2_neg(out1))
            if self.pass_samples:
                out2_neg = torch.cat((out2_neg, neg_sum_logs), dim=1).float()
            out3_neg = self.dense3_neg(out2_neg)
            out3_neg = self.sigmoid(out3_neg) if self.gates else out3_neg

        return out3_neg, out3_pos


class CombinedTargetHeadFromRnn(torch.nn.Module):
    def __init__(self, in_size=48, num_distr_classes_pos=4, num_distr_classes_neg=14, pos=True, neg=True, use_gates=True, pass_samples=True):
        super().__init__()

        self.pos, self.neg = pos, neg
        self.use_gates = use_gates
        self.dense = torch.nn.Linear(in_size, 256)
        self.distribution = DistributionTargetHead(256, num_distr_classes_pos, num_distr_classes_neg, pos, neg)
        self.regr_sums = RegressionTargetHead(256, False, pos, neg, pass_samples)
        self.regr_gates = RegressionTargetHead(256, True, pos, neg, pass_samples) if self.use_gates else None

        self.relu = torch.nn.ReLU()

    def forward(self, x):
        device = x[0].device if isinstance(x, tuple) else x.device
        sum_logs2 = torch.tensor(x[2][:, None], device=device) if isinstance(x, tuple) and len(x) > 2 else 0
        x, sum_logs1 = (x[0], torch.tensor(x[1][:, None], device=device)) if isinstance(x, tuple) else (x, 0)  # negative sum logs if len(x) == 3
        if not self.neg and self.pos:
            sum_logs1, sum_logs2 = sum_logs2, sum_logs1


        out1 = self.relu(self.dense(x))
        distr_neg, distr_pos = self.distribution(out1)

        sums_neg, sums_pos = self.regr_sums(out1, sum_logs1, sum_logs2)
        gate_neg, gate_pos = self.regr_gates(out1, sum_logs1, sum_logs2) if self.use_gates else (0, 0)

        return {'neg_sum': sum_logs1 * gate_neg + sums_neg * (1 - gate_neg),
                'neg_distribution': distr_neg,
                'pos_sum': sum_logs2 * gate_pos + sums_pos * (1 - gate_pos),
                'pos_distribution': distr_pos}


class TargetHeadFromAggFeatures(torch.nn.Module):
    def __init__(self, in_size=48, num_distr_classes=6):
        super().__init__()
        self.dense1 = torch.nn.Linear(in_size, 512)

        self.distribution = DistributionTargetHead(512, num_distr_classes)

        self.dense2_sums = torch.nn.Linear(512, 64)
        self.dense3_sums_neg = torch.nn.Linear(64, 1)
        self.dense3_sums_pos = torch.nn.Linear(64, 1)

        self.relu = torch.nn.ReLU()

    def forward(self, x):
        out1 = self.relu(self.dense1(x))

        distr_neg, distr_pos = self.distribution(out1)

        out2_sums = self.relu(self.dense2_sums(out1))
        out3_sums_neg = self.dense3_sums_neg(out2_sums)
        out3_sums_pos = self.dense3_sums_pos(out2_sums)

        return {'neg_sum': out3_sums_neg,
                'neg_distribution': distr_neg,
                'pos_sum': out3_sums_pos,
                'pos_distribution': distr_pos}


class DummyHead(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return {'neg_sum': x[0],
                'neg_distribution': x[1],
                'pos_sum': x[2],
                'pos_distribution': x[3]}


class GEGLU(torch.nn.Module):
    """
    References:
        Shazeer et al., "GLU Variants Improve Transformer," 2020.
        https://arxiv.org/abs/2002.05202

    Parameters
    ----------
    inputs
        Inputs to process [B, 2H]
    
    Returns
    -------
    outputs
        The outputs following the GEGLU activation [B, H]

    """
    
    def __init__(self):
        super().__init__()
        
    def geglu(self, x: torch.Tensor) -> torch.Tensor:
        assert x.shape[-1] % 2 == 0
        a, b = x.chunk(2, dim=-1)
        return a * F.gelu(b)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.geglu(x)


class StatPooling(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, pb: PaddedBatch):
        payload = pb.payload # B x T x H
        mask = pb.seq_len_mask.bool()
        #masked max
        inf = torch.zeros_like(mask, device=mask.device).float()
        inf[~mask] = -torch.inf
        pb_max = torch.max(payload + inf.unsqueeze(-1), dim=1)[0].squeeze()
        #masked min
        pb_min = torch.min(payload - inf.unsqueeze(-1), dim=1)[0].squeeze()
        #masked mean
        pb_mean = payload.sum(dim=1) / mask.float().sum(dim=1, keepdim=True)
        #masked std
        pb_std = payload - pb_mean.unsqueeze(1).expand_as(payload)
        pb_std = pb_std * mask.unsqueeze(-1)
        pb_std = torch.sqrt(torch.mean(torch.pow(pb_std,2), dim=1))

        out = torch.cat([pb_max, pb_min, pb_mean, pb_std], dim=1)
        return out


import numpy as np
import torch
from ptls.data_load.feature_dict import FeatureDict


class SeqLenLimit(FeatureDict):
    """
    This class is used as 'f_augmentation' argument for 
    ptls.data_load.datasets.augmentation_dataset.AugmentationDataset (AugmentationIterableDataset).
    """
    def __init__(self, max_seq_len, strategy='tail'):
        self.max_seq_len = max_seq_len
        self.strategy = strategy

        assert strategy in ('tail', 'head', 'random')

    def __call__(self, x):
        seq_len = self.get_seq_len(x)

        idx = self.get_idx(seq_len)
        new_x = self.seq_indexing(x, idx)
        return new_x

    def get_idx(self, seq_len):
        ix = np.arange(seq_len)
        if self.strategy == 'tail':
            # return ix[-self.max_seq_len:]
            return slice(-self.max_seq_len, None)
        elif self.strategy == 'head':
            # return ix[:self.max_seq_len]
            return slice(self.max_seq_len)
        elif self.strategy == 'random':
            if seq_len <= self.max_seq_len:
                return ix
            else:
                max_start_pos = seq_len - self.max_seq_len
                start_pos = np.random.choice(max_start_pos, 1)[0]
                return ix[start_pos:start_pos + self.max_seq_len]



import torch
from torch import nn as nn
from torch.nn import functional as F


class CPC_Loss(nn.Module):
    def __init__(self, n_negatives=None, n_forward_steps=None):
        super().__init__()
        self.n_negatives = n_negatives
        self.n_forward_steps = n_forward_steps

    def _get_preds(self, base_embeddings, mapped_ctx_embeddings):
        batch_size, max_seq_len, emb_size = base_embeddings.payload.shape
        _, _, _, n_forward_steps = mapped_ctx_embeddings.payload.shape
        seq_lens = mapped_ctx_embeddings.seq_lens
        device = mapped_ctx_embeddings.payload.device

        len_mask = torch.arange(max_seq_len).unsqueeze(0).expand(batch_size, -1).to(device)
        len_mask = (len_mask < seq_lens.unsqueeze(1).expand(-1, max_seq_len)).float()

        possible_negatives = base_embeddings.payload.view(batch_size * max_seq_len, emb_size)

        mask = len_mask.unsqueeze(0).expand(batch_size, *len_mask.shape).clone()

        mask = mask.reshape(batch_size, -1)
        sample_ids = torch.multinomial(mask, self.n_negatives)
        neg_samples = possible_negatives[sample_ids]

        positive_preds, neg_preds = [], []
        len_mask_exp = len_mask.unsqueeze(-1).unsqueeze(-1).to(device).expand(-1, -1, emb_size, n_forward_steps)
        trimmed_mce = mapped_ctx_embeddings.payload.mul(len_mask_exp)  # zero context vectors by sequence lengths
        for i in range(1, n_forward_steps + 1):
            ce_i = trimmed_mce[:, 0:max_seq_len - i, :, i - 1]
            be_i = base_embeddings.payload[:, i:max_seq_len]

            positive_pred_i = ce_i.mul(be_i).sum(axis=-1)
            positive_preds.append(positive_pred_i)

            neg_pred_i = ce_i.matmul(neg_samples.transpose(-2, -1))
            neg_preds.append(neg_pred_i)

        return positive_preds, neg_preds

    def forward(self, embeddings, _):
        base_embeddings, _, mapped_ctx_embeddings = embeddings
        device = mapped_ctx_embeddings.payload.device
        positive_preds, neg_preds = self._get_preds(base_embeddings, mapped_ctx_embeddings)

        step_losses = []
        for positive_pred_i, neg_pred_i in zip(positive_preds, neg_preds):
            step_loss = -F.log_softmax(torch.cat([positive_pred_i.unsqueeze(-1), neg_pred_i], dim=-1), dim=-1)[:, :, 0].mean()
            step_losses.append(step_loss)

        loss = torch.stack(step_losses).mean()
        return loss

    def cpc_accuracy(self, embeddings, _):
        base_embeddings, _, mapped_ctx_embeddings = embeddings
        positive_preds, neg_preds = self._get_preds(base_embeddings, mapped_ctx_embeddings)

        batch_size, max_seq_len, emb_size = base_embeddings.payload.shape
        seq_lens = mapped_ctx_embeddings.seq_lens
        device = mapped_ctx_embeddings.payload.device

        len_mask = torch.arange(max_seq_len).unsqueeze(0).expand(batch_size, -1).to(device)
        len_mask = (len_mask < seq_lens.unsqueeze(1).expand(-1, max_seq_len)).float()

        total, accurate = 0, 0
        for i, (positive_pred_i, neg_pred_i) in enumerate(zip(positive_preds, neg_preds)):
            i_mask = len_mask[:, (i + 1):max_seq_len].to(device)
            total += i_mask.sum().item()
            accurate += (((positive_pred_i.unsqueeze(-1).expand(*neg_pred_i.shape) > neg_pred_i) \
                          .sum(dim=-1) == self.n_negatives) * i_mask).sum().item()
        return accurate / total



import torch

from ptls.frames.coles.sampling_strategies.triplet_selector import TripletSelector


class RandomNegativeTripletSelector(TripletSelector):
    """
        Generate triplets with all positive pairs and random negative example for each anchor
    """

    def __init__(self, neg_count=1):
        super(RandomNegativeTripletSelector, self).__init__()
        self.neg_count = neg_count

    def get_triplets(self, embeddings, labels):
        n = labels.size(0)

        # construct matrix x, such as x_ij == 1 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        m = positive_pairs.size(0)
        anchor_labels = labels[positive_pairs[:, 0]]

        # construct matrix x (size m x n), such as x_ij == 1 <==> anchor_labels[i] == labels[j]
        x = (labels.expand(m, n) == anchor_labels.expand(n, m).t())

        negative_pairs = (x == 0).type(embeddings.dtype)
        negative_pairs_prob = (negative_pairs.t() / negative_pairs.sum(dim=1)).t()
        negative_pairs_rand = torch.multinomial(negative_pairs_prob, 1)

        triplets = torch.cat([positive_pairs, negative_pairs_rand], dim=1)

        return triplets


import logging
import warnings

import pytorch_lightning as pl
import numpy as np
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from sklearn.model_selection import train_test_split
import random
import glob

from ptls.data_load import padded_collate_emb_valid
from ptls.data_load.datasets.parquet_dataset import ParquetFiles, ParquetDataset


logger = logging.getLogger(__name__)


class EmbValidDataModule(pl.LightningDataModule):
    def __init__(self, type, setup, train, valid, pl_module, test=None):
        warnings.warn('Use `ptls.frames.PtlsDataModule` with `ptls.frames.supervised.SeqToTargetDataset` '
                      'or `ptls.frames.supervised.SeqToTargetIterableDataset`',
                      DeprecationWarning)
        super().__init__()

        self._type = type
        assert self._type in ('map', 'iterable')

        self.setup_conf = setup
        self.train_conf = train
        self.valid_conf = valid
        self.test_conf = test if test else self.valid_conf

        self.col_id = self.setup_conf.col_id
        self.col_id_dtype = {
            'str': str,
            'int': int,
        }[self.setup_conf.col_id_dtype]
        self.col_target = self.setup_conf.col_target
        self.category_names = pl_module.seq_encoder.category_names
        self.category_names.add('event_time')
        self.category_max_size = pl_module.seq_encoder.category_max_size

        self.train_dataset = None
        self.valid_dataset = None
        self.test_dataset = None
        self._train_targets = None
        self._valid_targets = None
        self._test_targets = None

        self._fold_info = None

    def prepare_data(self):
        if 'dataset_files' in self.setup_conf:
            self.setup_iterable_files()
        elif 'dataset_parts' in self.setup_conf:
            self.setup_iterable_parts()
        else:
            raise AttributeError(f'Missing dataset definition. One of `dataset_parts` or `dataset_files` expected')

    def setup_iterable_files(self):
       if (self.setup_conf.get('use_files_partially', None)):
           n_train = len(glob.glob(self.setup_conf.dataset_files.train_data_path + "/*.parquet"))
           ixes = list(range(n_train))
           train_ixes, test_ixes = train_test_split(ixes, test_size=int(n_train * (1 - self.setup_conf.train_part)), shuffle=True)
           train_data_files = ParquetFiles(self.setup_conf.dataset_files.train_data_path, train_ixes).data_files
           if self.setup_conf.same_file_for_test:
               test_ixes = random.sample(test_ixes, int(n_train * self.setup_conf.test_part))
               test_data_files = ParquetFiles(self.setup_conf.dataset_files.train_data_path, test_ixes).data_files
           else:
               n_test = len(glob.glob(self.setup_conf.dataset_files.test_data_path + "/*.parquet"))
               test_ixes = random.sample(ixes, int(n_test * self.setup_conf.test_part))
               test_data_files = ParquetFiles(self.setup_conf.dataset_files.test_data_path, test_ixes).data_files
       else:
           train_data_files = ParquetFiles(self.setup_conf.dataset_files.train_data_path).data_files
           test_data_files = ParquetFiles(self.setup_conf.dataset_files.test_data_path).data_files

       self.train_dataset = ParquetDataset(
           train_data_files,
           shuffle_files=True if self._type == 'iterable' else False
       )
       self.test_dataset = ParquetDataset(
           test_data_files,
           shuffle_files=False
       )
       self.test_dataset = list(tqdm(iter(self.test_dataset)))
       logger.info(f'Loaded {len(self.test_dataset)} for test')


    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=padded_collate_emb_valid,
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf.num_workers,
            batch_size=self.train_conf.batch_size,
            drop_last=self.train_conf.get('drop_last', False)
        )

    def test_dataloader(self):
         return DataLoader(
             dataset=self.test_dataset,
             collate_fn=padded_collate_emb_valid,
             num_workers=self.test_conf.num_workers,
             batch_size=self.test_conf.batch_size,
         )



import math
import torch

from ptls.frames.coles.metric import outer_cosine_similarity, outer_pairwise_distance, metric_recall_top_K, \
    BatchRecallTopK


def test_outer_cosine_similarity1():
    x = torch.eye(10, dtype=torch.float32)

    dists = outer_cosine_similarity(x, x)
    true_dists = torch.eye(10, dtype=torch.float32)

    assert torch.allclose(dists, true_dists)


def test_outer_cosine_similarity2():
    fi = 1
    x = torch.tensor([
        [math.cos(fi), -math.sin(fi)],
        [math.sin(fi), math.cos(fi)]
    ])

    dists = outer_cosine_similarity(x, x)
    true_dists = torch.eye(2, dtype=torch.float32)

    assert torch.allclose(dists, true_dists)


def test_outer_pairwise_distance1():
    x = torch.eye(10, dtype=torch.float32)

    dists = outer_pairwise_distance(x, x)
    true_dists = (1 - torch.eye(10, dtype=torch.float32)) * math.sqrt(2)

    assert torch.allclose(dists, true_dists, atol=1e-5)


def test_outer_pairwise_distance2():
    fi = 1
    x = torch.tensor([
        [math.cos(fi), -math.sin(fi)],
        [math.sin(fi), math.cos(fi)]
    ])

    dists = outer_pairwise_distance(x, x)
    true_dists = (1 - torch.eye(2, dtype=torch.float32)) * math.sqrt(2)

    assert torch.allclose(dists, true_dists, atol=1e-5)


def get_ml_data():
    b, c, h = 3, 2, 2  # Batch, num Classes, Hidden size
    x = torch.tensor([[0., 1.],       # 0
                      [0., 0.9],      # 0
                      [0.95, 0.],     # 0
                      [1., 0.],       # 1
                      [0.9, 0.],      # 1
                      [0.0, 0.95]])   # 1

    y = torch.arange(c).view(-1, 1).expand(c, b).reshape(-1)
    return x, y


def test_metric_recall_top_k():
    x, y = get_ml_data()
    metric = metric_recall_top_K(x, y, K=3 - 1, metric='euclidean')
    true_value = 1/3
    assert abs(metric - true_value) < 1e-6


def test_metric_recall_top_k_small_batch():
    x, y = get_ml_data()
    metric = metric_recall_top_K(x[1:5], y[1:5], K=3 - 1, metric='euclidean')
    true_value = 3 / 8
    assert abs(metric - true_value) < 1e-6


def test_batch_recall_top():
    x, y = get_ml_data()
    metric = BatchRecallTopK(K=3 - 1, metric='euclidean')
    metric(x, y)
    res = metric.compute()
    true_value = 1 / 3
    assert abs(res - true_value) < 1e-6


def test_batch_recall_top2_steps():
    x, y = get_ml_data()
    metric = BatchRecallTopK(K=3 - 1, metric='euclidean')
    metric(x, y)
    metric(x[1:5], y[1:5])
    res = metric.compute()
    true_value = 17 / 48
    assert abs(res - true_value) < 1e-6



import numpy as np
import torch
import pytorch_lightning as pl
from ptls.data_load import create_train_loader, create_validation_loader, TrxDataset
from ptls_tests.utils.data_generation import gen_trx_data


class RandomEventData(pl.LightningDataModule):
    def __init__(self, params, target_type='bin_cls'):
        super().__init__()
        self.hparams.update(params)
        self.target_type = target_type

    def train_dataloader(self):
        test_data = TrxDataset(
            gen_trx_data((torch.rand(1000) * 60 + 1).long(), target_type=self.target_type),
            y_dtype=np.int64 if self.target_type != 'regression' else np.float32,
        )
        train_loader = create_train_loader(test_data, self.hparams.train)
        return train_loader

    def val_dataloader(self):
        test_data = TrxDataset(
            gen_trx_data((torch.rand(200) * 60 + 1).long(), target_type=self.target_type),
            y_dtype=np.int64 if self.target_type != 'regression' else np.float32,
        )
        train_loader = create_validation_loader(test_data, self.hparams.valid)
        return train_loader



import torch
from numpy.testing import assert_almost_equal

from ptls.frames.coles.metric import outer_cosine_similarity


class HistogramLoss(torch.nn.Module):
    """
    HistogramLoss

    "Learning deep embeddings with histogram loss", NIPS 2016
    https://arxiv.org/abs/1611.00822
    code based on https://github.com/valerystrizh/pytorch-histogram-loss
    """

    def __init__(self, num_steps=100):
        super(HistogramLoss, self).__init__()
        self.step = 2 / (num_steps - 1)
        self.eps = 1 / num_steps
        self.t = torch.arange(-1, 1+self.step, self.step).view(-1, 1)
        self.tsize = self.t.size()[0]
        self.device = None

    def forward(self, embeddings, classes):
        def histogram(inds, size):

            s_repeat_ = s_repeat.clone()
            indsa = (s_repeat_floor - (self.t - self.step) > -self.eps) &  \
                    (s_repeat_floor - (self.t - self.step) < self.eps) & inds

            assert indsa.nonzero(as_tuple=False).size()[0] == size, 'Another number of bins should be used'
            zeros = torch.zeros((1, indsa.size()[1])).bool()
            zeros = zeros.to(self.device)
            indsb = torch.cat((indsa, zeros))[1:, :]
            s_repeat_[~(indsb | indsa)] = 0
            # indsa corresponds to the first condition of the second equation of the paper
            s_repeat_[indsa] = (s_repeat_ - self.t + self.step)[indsa] / self.step
            # indsb corresponds to the second condition of the second equation of the paper
            s_repeat_[indsb] = (-s_repeat_ + self.t + self.step)[indsb] / self.step

            return s_repeat_.sum(1) / size

        self.device = embeddings.device
        self.t = self.t.to(self.device)

        # L2 normalization
        classes_size = classes.size()[0]
        classes_eq = (classes.repeat(classes_size, 1)  == classes.view(-1, 1).repeat(1, classes_size)).data
        dists = outer_cosine_similarity(embeddings)

        assert ((dists > 1 + self.eps).sum().item() + (dists < -1 - self.eps).sum().item()) == 0, 'L2 normalization ' \
                                                                                                  'should be used '
        s_inds = torch.triu(torch.ones(classes_eq.size()), 1).bool()
        s_inds = s_inds.to(self.device)
        pos_inds = classes_eq[s_inds].repeat(self.tsize, 1)
        neg_inds = ~classes_eq[s_inds].repeat(self.tsize, 1)
        pos_size = classes_eq[s_inds].sum().item()
        neg_size = (~classes_eq[s_inds]).sum().item()
        s = dists[s_inds].view(1, -1)
        s = s.clamp(-1 + 1e-6, 1 - 1e-6)
        s_repeat = s.repeat(self.tsize, 1)
        s_repeat_floor = (torch.floor((s_repeat.data + 1.0 - 1e-6) / self.step) * self.step - 1.0).float()

        histogram_pos = histogram(pos_inds, pos_size)
        assert_almost_equal(histogram_pos.sum().item(), 1, decimal=1,
                            err_msg='Not good positive histogram', verbose=True)
        histogram_neg = histogram(neg_inds, neg_size)
        assert_almost_equal(histogram_neg.sum().item(), 1, decimal=1,
                            err_msg='Not good negative histogram', verbose=True)
        histogram_pos_repeat = histogram_pos.view(-1, 1).repeat(1, histogram_pos.size()[0])
        histogram_pos_inds = torch.tril(torch.ones(histogram_pos_repeat.size()), -1).bool()
        histogram_pos_inds = histogram_pos_inds.to(self.device)
        histogram_pos_repeat[histogram_pos_inds] = 0
        histogram_pos_cdf = histogram_pos_repeat.sum(0)
        loss = torch.sum(histogram_neg * histogram_pos_cdf)

        return loss


from typing import List

import pyspark
import pyspark.sql.functions as F
from pyspark.sql import Window

from ptls.preprocessing.base import ColTransformer
from ptls.preprocessing.pandas.col_transformer import ColTransformerPandasMixin


class UserGroupTransformer(ColTransformerPandasMixin, ColTransformer):
    """Groups transactions by user. Splits it by features.

    'event_time' column should be in dataset. We use it to order transactions

    Parameters
    ----------
    col_name_original:
        Column name with user_id - key for grouping
    cols_last_item:
        Only first value will be taken for these columns.
        All values as tensor will be taken for other columns
    max_trx_count:
        Keep only `max_trx_count` last transactions
    """
    def __init__(self,
                 col_name_original: str,
                 cols_last_item: List[str] = None,
                 max_trx_count: int = 10000,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target=None,
            is_drop_original_col=False,
        )
        self.cols_last_item = cols_last_item if cols_last_item is not None else []
        self.max_trx_count = max_trx_count

    def fit(self, x: pyspark.sql.DataFrame):
        # super().fit(x)
        if self.col_name_original not in x.columns:
            raise AttributeError(f'col_name_original="{self.col_name_original}" not in source dataframe. '
                                 f'Found {x.columns}')
        if 'event_time' not in x.columns:
            raise AttributeError(f'"event_time" not in source dataframe. '
                                 f'Found {x.columns}')
        return self


    def transform(self, x: pyspark.sql.DataFrame):
        col_list = [col for col in x.columns if col != self.col_name_original]

        df = x.withColumn('_rn', F.row_number().over(
            Window.partitionBy(self.col_name_original).orderBy(F.col('event_time').desc())))
        if self.max_trx_count is not None:
            df = df.where(F.col('_rn') <= self.max_trx_count)

        df = df.groupby(self.col_name_original).agg(
            *[
                F.sort_array(F.collect_list(F.struct('_rn', col)), asc=False).alias(col)
                for col in col_list if col not in self.cols_last_item
            ],
            *[
                F.struct(F.max(F.when(F.col('_rn') == 1, F.col(col))).alias(col)).alias(col)
                for col in col_list if col in self.cols_last_item
            ],
        )
        for col in col_list:
            df = df.withColumn(col, F.col(f'{col}.{col}'))

        # we don't heed to drop original column in this transformer
        # x = super().transform(x)
        return df



import pytorch_lightning as pl
import torch
from torchmetrics import MeanMetric

from ptls.frames.bert.losses.query_soft_max import QuerySoftmaxLoss
from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from ptls.nn import PBL2Norm
from ptls.data_load.padded_batch import PaddedBatch


class MLMPretrainModule(pl.LightningModule):
    """Masked Language Model (MLM) from [ROBERTA](https://arxiv.org/abs/1907.11692)

    Original sequence are encoded by `TrxEncoder`.
    Randomly sampled trx representations are replaced by MASK embedding.
    Transformer `seq_encoder` reconstruct masked embeddings.
    The loss function tends to make closer trx embedding and his predict.
    Negative samples are used to avoid trivial solution.

    Parameters
    ----------
    trx_encoder:
        Module for transform dict with feature sequences to sequence of transaction representations
    seq_encoder:
        Module for sequence processing. Generally this is transformer based encoder. Rnn is also possible
        Should works without sequence reduce
    hidden_size:
        Size of trx_encoder output.
    loss_temperature:
         temperature parameter of `QuerySoftmaxLoss`
    total_steps:
        total_steps expected in OneCycle lr scheduler
    max_lr:
        max_lr of OneCycle lr scheduler
    weight_decay:
        weight_decay of Adam optimizer
    pct_start:
        % of total_steps when lr increase
    norm_predict:
        use l2 norm for transformer output or not
    replace_proba:
        probability of masking transaction embedding
    neg_count:
        negative count for `QuerySoftmaxLoss`
    log_logits:
        if true than logits histogram will be logged. May be useful for `loss_temperature` tuning
    """

    def __init__(self,
                 trx_encoder: torch.nn.Module,
                 seq_encoder: AbsSeqEncoder,
                 total_steps: int,
                 hidden_size: int = None,
                 loss_temperature: float = 20.0,
                 max_lr: float = 0.001,
                 weight_decay: float = 0.0,
                 pct_start: float = 0.1,
                 norm_predict: bool = True,
                 replace_proba: float = 0.1,
                 neg_count: int = 1,
                 log_logits: bool = False,
                 ):

        super().__init__()
        self.save_hyperparameters(ignore=['trx_encoder', 'seq_encoder'])

        self.trx_encoder = trx_encoder
        self._seq_encoder = seq_encoder
        self._seq_encoder.is_reduce_sequence = False

        if self.hparams.norm_predict:
            self.fn_norm_predict = PBL2Norm()

        if hidden_size is None:
            hidden_size = trx_encoder.output_size

        self.token_mask = torch.nn.Parameter(torch.randn(1, 1, hidden_size), requires_grad=True)

        self.loss_fn = QuerySoftmaxLoss(temperature=loss_temperature, reduce=False)

        self.train_mlm_loss = MeanMetric()
        self.valid_mlm_loss = MeanMetric()

    def configure_optimizers(self):
        optim = torch.optim.Adam(self.parameters(),
                                 lr=self.hparams.max_lr,
                                 weight_decay=self.hparams.weight_decay,
                                 )
        scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer=optim,
            max_lr=self.hparams.max_lr,
            total_steps=self.hparams.total_steps,
            pct_start=self.hparams.pct_start,
            anneal_strategy='cos',
            cycle_momentum=False,
            div_factor=25.0,
            final_div_factor=10000.0,
            three_phase=False,
        )
        scheduler = {'scheduler': scheduler, 'interval': 'step'}
        return [optim], [scheduler]

    def get_mask(self, attention_mask):
        return torch.bernoulli(attention_mask.float() * self.hparams.replace_proba).bool()

    def mask_x(self, x, attention_mask, mask):
        shuffled_tokens = x[attention_mask.bool()]
        B, T, H = x.size()
        ix = torch.multinomial(torch.ones(shuffled_tokens.size(0)), B * T, replacement=True)
        shuffled_tokens = shuffled_tokens[ix].view(B, T, H)

        rand = torch.rand(B, T, device=x.device).unsqueeze(2).expand(B, T, H)
        replace_to = torch.where(
            rand < 0.8,
            self.token_mask.expand_as(x),  # [MASK] token 80%
            torch.where(
                rand < 0.9,
                shuffled_tokens,  # random token 10%
                x,  # unchanged 10%
            )
        )
        return torch.where(mask.bool().unsqueeze(2).expand_as(x), replace_to, x)

    def forward(self, z: PaddedBatch):
        out = self._seq_encoder(z)
        if self.hparams.norm_predict:
            out = self.fn_norm_predict(out)
        return out

    def get_neg_ix(self, mask):
        """Sample from predicts, where `mask == True`, without self element.
        sample from predicted tokens from batch
        """
        mask_num = mask.int().sum()
        mn = 1 - torch.eye(mask_num, device=mask.device)
        neg_ix = torch.multinomial(mn, self.hparams.neg_count)

        b_ix = torch.arange(mask.size(0), device=mask.device).view(-1, 1).expand_as(mask)[mask][neg_ix]
        t_ix = torch.arange(mask.size(1), device=mask.device).view(1, -1).expand_as(mask)[mask][neg_ix]
        return b_ix, t_ix

    def loss_mlm(self, x: PaddedBatch, is_train_step):
        mask = self.get_mask(x.seq_len_mask)
        masked_x = self.mask_x(x.payload, x.seq_len_mask, mask)

        out = self.forward(PaddedBatch(masked_x, x.seq_lens)).payload

        target = x.payload[mask].unsqueeze(1)  # N, 1, H
        predict = out[mask].unsqueeze(1)  # N, 1, H
        neg_ix = self.get_neg_ix(mask)
        negative = out[neg_ix[0], neg_ix[1]]  # N, nneg, H
        loss = self.loss_fn(target, predict, negative)

        if is_train_step and self.hparams.log_logits:
            with torch.no_grad():
                logits = self.loss_fn.get_logits(target, predict, negative)
            self.logger.experiment.add_histogram('mlm/logits',
                                                 logits.flatten().detach(), self.global_step)
        return loss

    def training_step(self, batch, batch_idx):
        x_trx = batch
        z_trx = self.trx_encoder(x_trx)  # PB: B, T, H
        loss_mlm = self.loss_mlm(z_trx, is_train_step=True)
        self.train_mlm_loss(loss_mlm)
        loss_mlm = loss_mlm.mean()
        self.log(f'mlm/loss', loss_mlm)
        return loss_mlm

    def validation_step(self, batch, batch_idx):
        x_trx = batch
        z_trx = self.trx_encoder(x_trx)  # PB: B, T, H
        loss_mlm = self.loss_mlm(z_trx, is_train_step=False)
        self.valid_mlm_loss(loss_mlm)

    def training_epoch_end(self, _):
        self.log(f'mlm/train_mlm_loss', self.train_mlm_loss, prog_bar=False)
        # self.train_mlm_loss reset not required here

    def validation_epoch_end(self, _):
        self.log(f'mlm/valid_mlm_loss', self.valid_mlm_loss, prog_bar=True)
        # self.valid_mlm_loss reset not required here



# coding: utf-8
import logging

import torch
from torch import nn as nn

from ptls.data_load.padded_batch import PaddedBatch

logger = logging.getLogger(__name__)


def projection_head(input_size, output_size):
    layers = [
        torch.nn.Linear(input_size, input_size),
        torch.nn.ReLU(),
        torch.nn.Linear(input_size, output_size),
    ]
    m = torch.nn.Sequential(*layers)
    return m


class ModelEmbeddingEnsemble(nn.Module):
    def __init__(self, submodels):
        super(ModelEmbeddingEnsemble, self).__init__()
        self.models = nn.ModuleList(submodels)

    def forward(self, x: PaddedBatch, h_0: torch.Tensor = None):
        """
        x - PaddedBatch of transactions sequences
        h_0 - previous state of embeddings (initial size for GRU). torch Tensor of shape (batch_size, embedding_size)
        """
        if h_0 is not None:
            h_0_splitted = torch.chunk(h_0, len(self.models), -1)
            out = torch.cat([m(x, h.contiguous()) for m, h in zip(self.models, h_0_splitted)], dim=-1)
        else:
            out = torch.cat([m(x) for i, m in enumerate(self.models)], dim=-1)
        return out


class ComplexModel(torch.nn.Module):
    def __init__(self, ml_model, params):
        super().__init__()
        self.ml_model = ml_model
        self.projection_ml_head = projection_head(params.rnn.hidden_size, params.ml_projection_head.output_size)
        self.projection_aug_head = torch.nn.Sequential(
            projection_head(params.rnn.hidden_size, params.aug_projection_head.output_size),
            torch.nn.LogSoftmax(dim=1)
        )

    def forward(self, x):
        encoder_output = self.ml_model(x)
        ml_head_output = self.projection_ml_head(encoder_output)
        aug_head_output = self.projection_aug_head(encoder_output)
        return aug_head_output, ml_head_output



import torch

from ptls.frames.coles.sampling_strategies.pair_selector import PairSelector


class AllPositivePairSelector(PairSelector):
    """
    Discards embeddings and generates all possible pairs given labels.
    If balance is True, negative pairs are a random sample to match the number of positive samples
    """

    def __init__(self, balance=True):
        super(AllPositivePairSelector, self).__init__()
        self.balance = balance

    def get_pairs(self, embeddings, labels):
        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        n = labels.size(0)
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)
        negative_pairs = torch.triu((x != 0).int(), diagonal=1).nonzero(as_tuple=False)

        if self.balance:
            negative_pairs = negative_pairs[torch.randperm(len(negative_pairs))[:len(positive_pairs)]]

        return positive_pairs, negative_pairs


import torch
import numpy as np

from ptls.loss import DistributionTargetsLoss


def test_best_loss():
    eps = 1e-7

    prediction = {'neg_sum': torch.tensor([[np.log(10 + 1)]]), 
                  'neg_distribution': torch.tensor([[100., 0., 0., 0., 0., 0.]]),
                  'pos_sum': torch.tensor([[0]]),
                  'pos_distribution': torch.tensor([[0., 100., 0., 0., 0., 0.]])}

    label = {'neg_sum': np.array([[10]]),
             'neg_distribution': np.array([[1., 0., 0., 0., 0., 0.]]),
             'pos_sum': np.array([[0]]),
             'pos_distribution': np.array([[0., 1., 0., 0., 0., 0.]])}

    loss = DistributionTargetsLoss()
    out = loss(prediction, label)
    assert abs(out.item() - 0.) < eps
    assert type(out) is torch.Tensor


def test_loss_300():
    eps = 1e-7
    
    prediction = {'neg_sum': torch.tensor([[10]]), 
                  'neg_distribution': torch.tensor([[100., 0., 0., 0., 0., 0.]]),
                  'pos_sum': torch.tensor([[0]]),
                  'pos_distribution': torch.tensor([[0., 100., 0., 0., 0., 0.]])}

    label = {'neg_sum': np.array([[0]]),
             'neg_distribution': np.array([[1., 0., 0., 0., 0., 0.]]),
             'pos_sum': np.array([[0]]),
             'pos_distribution': np.array([[0., 1., 0., 0., 0., 0.]])}

    loss = DistributionTargetsLoss()
    out = loss(prediction, label)
    assert abs(out.item() - 300.) < eps
    assert type(out) is torch.Tensor

    
def test_usual_loss_first():
    eps = 1e-7

    prediction = {'neg_sum': torch.tensor([[-1.]]), 
                  'neg_distribution': torch.tensor([[0.1, 0.2, 0.1, 0.1, 0.3, 0.2]]),
                  'pos_sum': torch.tensor([[ 1.]]),
                  'pos_distribution': torch.tensor([[0.1, 0.2, 0.1, 0.1, 0.3, 0.2]])}

    label = {'neg_sum': np.array([[-1.]]),
             'neg_distribution': np.array([[0.1, 0.2, 0.1, 0.1, 0.3, 0.2]]),
             'pos_sum': np.array([[1.]]),
             'pos_distribution': np.array([[0.1, 0.2, 0.1, 0.1, 0.3, 0.2]])}

    loss = DistributionTargetsLoss()
    out = loss(prediction, label)

    assert abs(out.item() - 12.138458251953125) < eps
    assert type(out) is torch.Tensor


def test_usual_loss_second():
    eps = 1e-7
    
    prediction = {'neg_sum': torch.tensor([[-1.]]), 
                  'neg_distribution': torch.tensor([[0.1, 0.2, 0.1, 0.1, 0.3, 0.2]]),
                  'pos_sum': torch.tensor([[ 1.]]),
                  'pos_distribution': torch.tensor([[0.3, 0.5, 0., 0.1, 0.1, 0.0]])}

    label = {'neg_sum': np.array([[-10.]]),
             'neg_distribution': np.array([[0.5, 0.5, 0.0, 0.0, 0.0, 0.0]]),
             'pos_sum': np.array([[8.]]),
             'pos_distribution': np.array([[0.1, 0.1, 0.1, 0.1, 0.1, 0.5]])}

    loss = DistributionTargetsLoss()
    out = loss(prediction, label)

    assert abs(out.item() - 38.563011169433594) < eps
    assert type(out) is torch.Tensor

    
def test_one_class():
    eps = 1e-7

    prediction = {'neg_sum': torch.tensor([[-1.]]), 
                  'neg_distribution': torch.tensor([[1., 0., 0., 0., 0., 0.]]),
                  'pos_sum': torch.tensor([[ 1.]]),
                  'pos_distribution': torch.tensor([[0., 1., 0., 0., 0., 0.]])}

    label = {'neg_sum': np.array([[-1.]]),
             'neg_distribution': np.array([[1., 0., 0., 0., 0., 0.]]),
             'pos_sum': np.array([[1.]]),
             'pos_distribution': np.array([[0., 1., 0., 0., 0., 0.]])}

    loss = DistributionTargetsLoss()
    out = loss(prediction, label)

    assert abs(out.item() - 10.703149795532227) < eps
    assert type(out) is torch.Tensor



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class TargetExtractor(IterableProcessingDataset):
    """Deprecated. Use SeqToTargetDataset

    """
    def __init__(self, target_col, drop_from_features=True):
        """Extract value from `target_col` and mention it as `y`

        for x, * in seq:
            y = x[target_col]
            yield x, y

        Args:
            target_col: field where `y` is stored

        """
        super().__init__()

        self._target_col = target_col
        self._drop_from_features = drop_from_features

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            y = features[self._target_col]
            if self._drop_from_features:
                features = {k: v for k, v in features.items() if k != self._target_col}
            yield features, y


class FakeTarget(IterableProcessingDataset):
    """Deprecated. Only single dict allowed

    """
    def __init__(self):
        """Create target equal 0 (for consistency)

        for x in seq:
            yield x, 0

        """
        super().__init__()

    def __iter__(self):
        for rec in self._src:
            yield rec, 0



import torch
from ptls.data_load import AugmentationChain


class AugmentationDataset(torch.utils.data.Dataset):
    def __init__(self, data, f_augmentations=None):
        self.data = data
        self.f_augmentations = AugmentationChain(f_augmentations)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        rec = self.data[item]
        return self.f_augmentations(rec)
    

class AugmentationIterableDataset(torch.utils.data.IterableDataset):
    def __init__(self, data, f_augmentations=None):
        assert isinstance(data, torch.utils.data.IterableDataset), " 'data' argument must be torch.utils.data.IterableDataset"
        self.data = data
        self.f_augmentations = AugmentationChain(f_augmentations)
    
    def __iter__(self):
        for rec in self.data:
            yield self.f_augmentations(rec)



import numpy as np
import pandas as pd

from ptls.data_load.iterable_processing.id_filter_df import IdFilterDf


def test_df_filter_one_col():
    i_filter = IdFilterDf(pd.DataFrame({
        'col1': [8, 9, 1, 2, 3],
    }))

    data = [{'col1': i, 'feature_1': np.random.randint(1, 10, 100)} for i in range(20)]
    data = i_filter(data)
    data = [x['col1'] for x in data]
    assert data == [1, 2, 3, 8, 9]


def test_df_filter_two_col():
    i_filter = IdFilterDf(pd.DataFrame({
        'col1': [8, 9, 1, 2, 3],
        'col2': [1, 1, 1, 2, 2],
    }))

    data = [{'col1': i, 'col2': 1, 'feature_1': np.random.randint(1, 10, 100)} for i in range(20)]
    data = i_filter(data)
    data = [x['col1'] for x in data]
    assert data == [1, 8, 9]


def test_df_filter_two_col_type_cast_to_int():
    i_filter = IdFilterDf(pd.DataFrame({
        'col1': [8, 9, 1, 2, 3],
        'col2': [1, 1, 1, 2, 2],
    }))

    data = [{'col1': i, 'col2': '1', 'feature_1': np.random.randint(1, 10, 100)} for i in range(20)]
    data = i_filter(data)
    data = [x['col1'] for x in data]
    assert data == [1, 8, 9]


def test_df_filter_two_col_type_cast_to_datetime():
    i_filter = IdFilterDf(pd.DataFrame({
        'col1': [8, 9, 1, 2, 3],
        'col2': pd.to_datetime([
            '2012-01-08 00:00:00',
            '2012-01-08 00:00:00',
            '2012-01-08 00:00:00',
            '2020-01-08 00:00:00',
            '2020-01-08 00:00:00',
        ]),
    }))

    data = [{'col1': i, 'col2': '2012-01-08', 'feature_1': np.random.randint(1, 10, 100)} for i in range(20)]
    data = i_filter(data)
    data = [x['col1'] for x in data]
    assert data == [1, 8, 9]


def test_df_filter_no_relevant():
    i_filter = IdFilterDf(pd.DataFrame({
        'col1': [8, 9, 1, 2, 3],
        'col2': pd.to_datetime([
            '2012-01-08 00:00:00',
            '2012-02-08 00:00:00',
            '2012-03-08 00:00:00',
            '2012-04-08 00:00:00',
            '2012-05-08 00:00:00',
        ]),
    }))

    data = [{'col1': i, 'col2': '2012-01-03', 'feature_1': np.random.randint(1, 10, 100)} for i in range(20)]
    data = i_filter(data)
    data = [x['col1'] for x in data]
    assert data == []



import torch
from ptls.data_load.feature_dict import FeatureDict


class AllTimeShuffle(FeatureDict):
    """
    Shuffle all transactions in event sequence.
    This class is used as 'f_augmentation' argument for 
    ptls.data_load.datasets.augmentation_dataset.AugmentationDataset (AugmentationIterableDataset).
    """
    def __init__(self, event_time_name='event_time'):
        self.event_time_name = event_time_name

    @staticmethod
    def get_perm_ix(event_time):
        n = len(event_time)
        return torch.randperm(n)

    def __call__(self, x):
        ix = self.get_perm_ix(x[self.event_time_name])
        new_x = self.seq_indexing(x, ix)
        return new_x



import numpy as np
import pytest
import torch

from ptls.data_load.utils import collate_feature_dict


def test_collate_feature_dict():
    batch = [
        {
            'bin': 0,
            'target_bin': 2,
            'pp': 0.6,
            'user_id': 'a',
            'lists': [1, 2, 3],
            'mcc': torch.tensor([1, 2, 3, 4, 5]),
            'event_time': torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5]),
            'target_array': torch.tensor([0.1, 0.2, 0.3]),
        },
        {
            'bin': 4,
            'target_bin': 3,
            'pp': 0.2,
            'user_id': 'b',
            'lists': [4, 1, 2, 3],
            'mcc': torch.tensor([1, 2]),
            'event_time': torch.tensor([0.3, 0.4]),
            'target_array': torch.tensor([0.2, 0.1, 0.3]),
        },
        {
            'bin': 5,
            'target_bin': 0,
            'pp': 0.1,
            'user_id': 'c',
            'lists': [3, 1, 0],
            'mcc': torch.tensor([1, 3, 4, 5]),
            'event_time': torch.tensor([0.2, 0.3, 0.4, 0.5]),
            'target_array': torch.tensor([0.5, 0.2, 0.3]),
        },
    ]
    pb = collate_feature_dict(batch)
    torch.testing.assert_close(pb.seq_lens, torch.LongTensor((5, 2, 4)))
    torch.testing.assert_close(pb.payload['bin'], torch.LongTensor([0, 4, 5]))
    torch.testing.assert_close(pb.payload['target_bin'], torch.LongTensor([2, 3, 0]))
    torch.testing.assert_close(pb.payload['pp'], torch.FloatTensor([0.6, 0.2, 0.1]))
    np.testing.assert_equal(pb.payload['user_id'], np.array(['a', 'b', 'c']))
    assert type(pb.payload['lists']) is np.ndarray
    assert pb.payload['lists'].dtype.kind == 'O'
    mcc_expected = torch.LongTensor([
        [1, 2, 3, 4, 5],
        [1, 2, 0, 0, 0],
        [1, 3, 4, 5, 0],
    ])
    torch.testing.assert_close(pb.payload['mcc'], mcc_expected)
    event_time_expected = torch.FloatTensor([
        [0.1, 0.2, 0.3, 0.4, 0.5],
        [0.3, 0.4, 0.0, 0.0, 0.0],
        [0.2, 0.3, 0.4, 0.5, 0.0],
    ])
    torch.testing.assert_close(pb.payload['event_time'], event_time_expected)
    target_array_expected = torch.FloatTensor([
        [0.1, 0.2, 0.3],
        [0.2, 0.1, 0.3],
        [0.5, 0.2, 0.3],
    ])
    torch.testing.assert_close(pb.payload['target_array'], target_array_expected)


def test_collate_feature_dict_inconsistent_new():
    batch = [
        {
            'mcc': torch.tensor([1, 2, 3, 4, 5]),
            'amount': torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5]),
            'target_float': 0.6,
        },
        {
            'mcc': torch.tensor([1, 2, 3]),
            'amount': torch.tensor([0.1, 0.2, 0.3]),
            'target_int': 0,
            'target_float': 0.5,
        },
    ]
    with pytest.raises(AssertionError):
        pb = collate_feature_dict(batch)

def test_collate_feature_dict_inconsistent_missing():
    batch = [
        {
            'mcc': torch.tensor([1, 2, 3, 4, 5]),
            'amount': torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5]),
            'target_int': 0,
            'target_float': 0.5,
        },
        {
            'mcc': torch.tensor([1, 2, 3]),
            'amount': torch.tensor([0.1, 0.2, 0.3]),
            'target_int': 0,
        },
    ]
    with pytest.raises(AssertionError):
        _ = collate_feature_dict(batch)



import torch
import numpy as np
from torch import nn as nn

from ptls.data_load.padded_batch import PaddedBatch


class TimeStepShuffle(nn.Module):
    def forward(self, x: PaddedBatch):
        shuffled = []
        for seq, slen in zip(x.payload, x.seq_lens):
            idx = torch.randperm(slen) + 1
            pad_idx = torch.arange(slen + 1, len(seq))
            idx = torch.cat([torch.zeros(1, dtype=torch.long), idx, pad_idx])
            shuffled.append(seq[idx])

        shuffled = PaddedBatch(torch.stack(shuffled), x.seq_lens)
        return shuffled


class LastStepEncoder(nn.Module):
    """
    Class is used by ptls.nn.RnnSeqEncoder for reducing RNN output with shape (B, L, H), where
        B - batch size
        L - sequence length
        H - hidden RNN size
    to embeddings tensor with shape (B, H). The last hidden state is used for embedding.
    
    Example of usage: seq_encoder = RnnSeqEncoder(..., reducer='last_step')
    """
    def forward(self, x: PaddedBatch):
        h = x.payload[range(len(x.payload)), [l - 1 for l in x.seq_lens]]
        return h


class FirstStepEncoder(nn.Module):
    """
    Class is used by ptls.nn.RnnSeqEncoder class for reducing RNN output with shape (B, L, H)
    to embeddings tensor with shape (B, H). The first hidden state is used for embedding.
    
    where:
        B - batch size
        L - sequence length
        H - hidden RNN size
    
    Example of usage: seq_encoder = RnnSeqEncoder(..., reducer='first_step')
    """
    def forward(self, x: PaddedBatch):
        h = x.payload[:, 0, :]  # [B, L, H] -> [B, H]
        return h
    

class LastMaxAvgEncoder(nn.Module):
    """
    Class is used by ptls.nn.RnnSeqEncoder class for reducing RNN output with shape (B, L, H)
    to embeddings tensor with shape (B, 3 * H). Embeddings are created by concatenating:
        - last hidden state from RNN output,
        - max pool over all hidden states of RNN output,
        - average pool over all hidden states of RNN output.
        
    where:
        B - batch size
        L - sequence length
        H - hidden RNN size
        
    Example of usage: seq_encoder = RnnSeqEncoder(..., reducer='last_max_avg')
    """
    def forward(self, x: PaddedBatch):
        rnn_max_pool = x.payload.max(dim=1)[0]
        rnn_avg_pool = x.payload.sum(dim=1) / x.seq_lens.unsqueeze(-1)
        h = x.payload[range(len(x.payload)), [l - 1 for l in x.seq_lens]]
        h = torch.cat((h, rnn_max_pool, rnn_avg_pool), dim=-1)
        return h

    
class SkipStepEncoder(nn.Module):
    def __init__(self, step_size):
        super().__init__()
        self.step_size = step_size

    def forward(self, x: PaddedBatch):
        max_len = x.payload.shape[1] - 1
        s = self.step_size

        first_dim_idx = []
        second_dim_idx = []
        for i, l in enumerate(x.seq_lens):
            idx_to_take = np.arange(min(l - 1, s - 1 + l % s), l, s)
            pad_idx = np.array([max_len - 1] * (max_len // s - len(idx_to_take)), dtype=np.int32)
            idx_to_take = np.concatenate([[-1], idx_to_take, pad_idx]) + 1
            first_dim_idx.append(np.ones(len(idx_to_take)) * i)
            second_dim_idx.append(idx_to_take)

        out = x.payload[first_dim_idx, second_dim_idx]
        out_lens = torch.tensor([min(1, l // self.step_size) for l in x.seq_lens])

        return PaddedBatch(out, out_lens)



import pyspark.sql.functions as F

from ptls.preprocessing.base import ColTransformer
from ptls.preprocessing.pyspark.col_transformer import ColTransformerPysparkMixin


class ColIdentityEncoder(ColTransformerPysparkMixin, ColTransformer):
    """Dont change original column

    May rename or copy original columns.

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Target column name. Transformed column will be placed here
        If `col_name_target is None` then original column will be replaced by transformed values.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.

    """
    def transform(self, x):
        x = x.withColumn(self.col_name_target, F.col(self.col_name_original))
        x = super().transform(x)
        return x



from .rnn_encoder import RnnEncoder
from .transformer_encoder import TransformerEncoder
from .longformer_encoder import LongformerEncoder
from .gpt_encoder import GptEncoder
from .custom_encoder import Encoder

from .containers import RnnSeqEncoder, TransformerSeqEncoder, LongformerSeqEncoder, CustomSeqEncoder
from .agg_feature_seq_encoder import AggFeatureSeqEncoder

# from ptls.nn.seq_encoder.rnn_seq_encoder_distribution_target import RnnSeqEncoderDistributionTarget
# from ptls.nn.seq_encoder.statistics_encoder import StatisticsEncoder



import torch

from ptls.data_load.augmentations.all_time_shuffle import AllTimeShuffle


def test_shuffle():
    i_filter = AllTimeShuffle()
    data = {'event_time': torch.arange(5), 'mcc': torch.arange(5)}
    data = i_filter(data)
    assert len(data['event_time']) == 5
    assert data['mcc'].sum() == 10


def test_shuffle_with_target():
    i_filter = AllTimeShuffle()
    data = {'event_time': torch.arange(5), 'mcc': torch.arange(5), 'target': 0}
    data = i_filter(data)
    assert len(data['event_time']) == 5
    assert data['mcc'].sum() == 10



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset
from ptls.data_load.augmentations.seq_len_limit import SeqLenLimit


class ISeqLenLimit(IterableProcessingDataset):
    def __init__(self, max_seq_len, strategy='tail'):
        super().__init__()

        self.proc = SeqLenLimit(max_seq_len, strategy)

    def process(self, features):
        return self.proc(features)



import logging

import torch
import numpy as np
import torchmetrics


logger = logging.getLogger(__name__)


def outer_pairwise_distance(A, B=None):
    """
        Compute pairwise_distance of Tensors
            A (size(A) = n x d, where n - rows count, d - vector size) and
            B (size(A) = m x d, where m - rows count, d - vector size)
        return matrix C (size n x m), such as C_ij = distance(i-th row matrix A, j-th row matrix B)

        if only one Tensor was given, computer pairwise distance to itself (B = A)
    """

    if B is None: B = A

    max_size = 2 ** 26
    n = A.size(0)
    m = B.size(0)
    d = A.size(1)

    if n * m * d <= max_size or m == 1:

        return torch.pairwise_distance(
            A[:, None].expand(n, m, d).reshape((-1, d)),
            B.expand(n, m, d).reshape((-1, d))
        ).reshape((n, m))

    else:

        batch_size = max(1, max_size // (n * d))
        batch_results = []
        for i in range((m - 1) // batch_size + 1):
            id_left = i * batch_size
            id_rigth = min((i + 1) * batch_size, m)
            batch_results.append(outer_pairwise_distance(A, B[id_left:id_rigth]))

        return torch.cat(batch_results, dim=1)


def outer_cosine_similarity(A, B=None):
    """
        Compute cosine_similarity of Tensors
            A (size(A) = n x d, where n - rows count, d - vector size) and
            B (size(A) = m x d, where m - rows count, d - vector size)
        return matrix C (size n x m), such as C_ij = cosine_similarity(i-th row matrix A, j-th row matrix B)

        if only one Tensor was given, computer pairwise distance to itself (B = A)
    """

    if B is None: B = A

    max_size = 2 ** 32
    n = A.size(0)
    m = B.size(0)
    d = A.size(1)

    if n * m * d <= max_size or m == 1:

        A_norm = torch.div(A.transpose(0, 1), A.norm(dim=1)).transpose(0, 1)
        B_norm = torch.div(B.transpose(0, 1), B.norm(dim=1)).transpose(0, 1)
        return torch.mm(A_norm, B_norm.transpose(0, 1))

    else:

        batch_size = max(1, max_size // (n * d))
        batch_results = []
        for i in range((m - 1) // batch_size + 1):
            id_left = i * batch_size
            id_rigth = min((i + 1) * batch_size, m)
            batch_results.append(outer_cosine_similarity(A, B[id_left:id_rigth]))

        return torch.cat(batch_results, dim=1)


def metric_recall_top_K(X, y, K, metric='cosine'):
    """
        calculate metric R@K
        X - tensor with size n x d, where n - number of examples, d - size of embedding vectors
        y - true labels
        N - count of closest examples, which we consider for recall calcualtion
        metric: 'cosine' / 'euclidean'.
            !!! 'euclidean' - to slow for datasets bigger than 100K rows
    """
    # TODO: take K from `y`
    res = []

    n = X.size(0)
    d = X.size(1)
    max_size = 2 ** 32
    batch_size = max(1, max_size // (n * d))

    with torch.no_grad():

        for i in range(1 + (len(X) - 1) // batch_size):

            id_left = i * batch_size
            id_right = min((i + 1) * batch_size, len(y))
            y_batch = y[id_left:id_right]

            if metric == 'cosine':
                pdist = -1 * outer_cosine_similarity(X, X[id_left:id_right])
            elif metric == 'euclidean':
                pdist = outer_pairwise_distance(X, X[id_left:id_right])
            else:
                raise AttributeError(f'wrong metric "{metric}"')

            values, indices = pdist.topk(K + 1, 0, largest=False)

            y_rep = y_batch.repeat(K, 1)
            res.append((y[indices[1:]] == y_rep).sum().item())

    return np.sum(res) / len(y) / K


class BatchRecallTopK(torchmetrics.MeanMetric):
    def __init__(self, K, metric='cosine'):
        super().__init__()

        self.k = K
        self.metric = metric

    def update(self, preds, target):
        super().update(metric_recall_top_K(preds, target, self.k, self.metric))



# Pair
from .all_positive_pair_selector import AllPositivePairSelector
from .hard_negative_pair_selector import HardNegativePairSelector
from .distance_weighted_pair_selector import DistanceWeightedPairSelector
from .pairwise_matrix_selector import PairwiseMatrixSelector

# Triplets
from .all_triplets_selector import AllTripletSelector
from .hard_triplet_selector import HardTripletSelector
from .random_negative_triplet_selector import RandomNegativeTripletSelector
from .semi_hard_triplet_selector import SemiHardTripletSelector

# Masker
from .matrix_masker import MatrixMasker


import torch
from ptls.frames.supervised.seq_to_target_dataset import SeqToTargetDataset


def test_seq_to_target_dataset_default_long():
    dataset = SeqToTargetDataset([{
        'mcc_code': torch.randint(1, 10, (seq_len,)),
        'amount': torch.randn(seq_len),
        'event_time': torch.arange(seq_len),  # shows order between transactions
        'target': target,
    } for seq_len, target in zip(
        torch.randint(100, 200, (4,)),
        [0, 0, 1, 1],
    )], target_col_name='target')
    dl = torch.utils.data.DataLoader(dataset, batch_size=10, collate_fn=dataset.collate_fn)
    x, y = next(iter(dl))
    torch.testing.assert_close(y, torch.LongTensor([0, 0, 1, 1]))

def test_seq_to_target_dataset_default_double():
    dataset = SeqToTargetDataset([{
        'mcc_code': torch.randint(1, 10, (seq_len,)),
        'amount': torch.randn(seq_len),
        'event_time': torch.arange(seq_len),  # shows order between transactions
        'target': target,
    } for seq_len, target in zip(
        torch.randint(100, 200, (4,)),
        [0.1, 0.4, 1.0, 0.9],
    )], target_col_name='target')
    dl = torch.utils.data.DataLoader(dataset, batch_size=10, collate_fn=dataset.collate_fn)
    x, y = next(iter(dl))
    torch.testing.assert_close(y, torch.FloatTensor([0.1, 0.4, 1.0, 0.9]))


def test_seq_to_target_dataset_type_double():
    dataset = SeqToTargetDataset([{
        'mcc_code': torch.randint(1, 10, (seq_len,)),
        'amount': torch.randn(seq_len),
        'event_time': torch.arange(seq_len),  # shows order between transactions
        'target': target,
    } for seq_len, target in zip(
        torch.randint(100, 200, (4,)),
        [0, 0, 1, 1],
    )], target_col_name='target', target_dtype=torch.double)
    dl = torch.utils.data.DataLoader(dataset, batch_size=10, collate_fn=dataset.collate_fn)
    x, y = next(iter(dl))
    torch.testing.assert_close(y, torch.DoubleTensor([0.0, 0.0, 1.0, 1.0]))

def test_seq_to_target_dataset_type_array():
    dataset = SeqToTargetDataset([{
        'mcc_code': torch.randint(1, 10, (seq_len,)),
        'amount': torch.randn(seq_len),
        'event_time': torch.arange(seq_len),  # shows order between transactions
        'target': torch.randn(2, 8),
    } for seq_len in torch.randint(100, 200, (1000,))], target_col_name='target')
    dl = torch.utils.data.DataLoader(dataset, batch_size=10, collate_fn=dataset.collate_fn)
    x, y = next(iter(dl))
    assert y.size() == (10, 2, 8)
    assert y.dtype == torch.float32



import torch
from torch.nn import functional as F

from ptls.frames.coles.sampling_strategies.triplet_selector import TripletSelector
from ptls.frames.coles.metric import outer_pairwise_distance


class SemiHardTripletSelector(TripletSelector):
    """
        Generate triplets with semihard sampling strategy

        "FaceNet: A Unified Embedding for Face Recognition and Clustering", CVPR 2015
        https://arxiv.org/abs/1503.03832
    """

    def __init__(self, neg_count=1):
        super(SemiHardTripletSelector, self).__init__()
        self.neg_count = neg_count

    def get_triplets(self, embeddings, labels):
        n = labels.size(0)

        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        m = positive_pairs.size(0)

        anchor_embed = embeddings[positive_pairs[:, 0]].detach()
        anchor_labels = labels[positive_pairs[:, 0]]

        pos_embed = embeddings[positive_pairs[:, 1]].detach()

        D_ap = F.pairwise_distance(anchor_embed, pos_embed)

        # construct matrix x (size m x n), such as x_ij == 1 <==> anchor_labels[i] == labels[j]
        x = (labels.expand(m, n) == anchor_labels.expand(n, m).t())

        mat_distances = outer_pairwise_distance(anchor_embed, embeddings.detach())  # pairwise_distance anchors x all

        neg_mat_distances = mat_distances * (x == 0).type(mat_distances.dtype)  # filter: get only negative pairs

        # negatives_outside: smallest D_an where D_an > D_ap.
        upper_bound = int((2 * n) ** 0.5) + 1
        negatives_outside = (upper_bound - neg_mat_distances) * \
                            (neg_mat_distances > D_ap.expand(n, m).t()).type(neg_mat_distances.dtype)
        values, negatives_outside = negatives_outside.topk(k=1, dim=1, largest=True)

        # negatives_inside: largest D_an
        values, negatives_inside = neg_mat_distances.topk(k=1, dim=1, largest=True)

        # whether exist negative n, such that D_an > D_ap.
        semihard_exist = ((neg_mat_distances > D_ap.expand(n, m).t()).sum(dim=1) > 0).view(-1, 1)

        negatives_indeces = torch.where(semihard_exist, negatives_outside, negatives_inside)

        triplets = torch.cat([positive_pairs, negatives_indeces], dim=1)

        return triplets


from ptls.util import ListSubset


def test_list_subset_iter():
    ls = ListSubset([1, 2, 3], [0, 2])
    res = [e for e in ls]
    assert [1, 3] == res


def test_list_subset_geitem():
    ls = ListSubset([1, 2, 3], [0, 2])
    assert len(ls) == 2
    assert ls[1] == 3



from .category_size_clip import CategorySizeClip
from .feature_rename import FeatureRename
from .feature_bin_scaler import FeatureBinScaler
from .feature_filter import FeatureFilter
from .feature_type_cast import FeatureTypeCast
from .filter_non_array import FilterNonArray
from .id_filter import IdFilter
from .id_filter_df import IdFilterDf
from .iterable_seq_len_limit import ISeqLenLimit
from .iterable_shuffle import IterableShuffle
from .seq_len_filter import SeqLenFilter
from .take_first_trx import TakeFirstTrx
from .target_empty_filter import TargetEmptyFilter
from .target_extractor import TargetExtractor
from .target_join import TargetJoin
from .target_move import TargetMove
from .to_torch_tensor import ToTorch



from torch import nn as nn
from torch.nn import functional as F


class TripletLoss(nn.Module):
    """
    Triplets loss

    "Deep metric learning using triplet network", SIMBAD 2015
    https://arxiv.org/abs/1412.6622
    """

    def __init__(self, margin, triplet_selector):
        super(TripletLoss, self).__init__()
        self.margin = margin
        self.triplet_selector = triplet_selector

    def forward(self, embeddings, target):

        triplets = self.triplet_selector.get_triplets(embeddings, target)

        if embeddings.is_cuda:
            triplets = triplets.cuda()

        ap_distances = F.pairwise_distance(embeddings[triplets[:, 0]], embeddings[triplets[:, 1]])
        an_distances = F.pairwise_distance(embeddings[triplets[:, 0]], embeddings[triplets[:, 2]])
        losses = F.relu(ap_distances - an_distances + self.margin)

        return losses.sum()


import torch


class ComplexLoss(torch.nn.Module):
    """Works like `ptls.loss.MultiLoss`

    """
    def __init__(self, ml_loss, aug_loss, ml_loss_weight=1.):
        super(ComplexLoss, self).__init__()
        self.aug_loss = aug_loss
        self.ml_loss = ml_loss
        self.ml_loss_weight = ml_loss_weight

    def forward(self, model_outputs, target):
        aug_output, ml_output = model_outputs
        aug_target = target[:, 0]
        ml_target = target[:, 1]
        aug = self.aug_loss(aug_output, aug_target) * (1 - self.ml_loss_weight)
        ml = self.ml_loss(ml_output, ml_target) * self.ml_loss_weight
        return aug + ml



import numpy as np

from ptls.nn.seq_encoder import RnnSeqEncoder


class RnnSeqEncoderDistributionTarget(RnnSeqEncoder):
    def transform(self, x):
        return np.sign(x) * np.log(np.abs(x) + 1)

    def transform_inv(self, x):
        return np.sign(x) * (np.exp(np.abs(x)) - 1)

    def __init__(self,
                 trx_encoder,
                 hidden_size,
                 type,
                 bidir,
                 trainable_starter,
                 head_layers,
                 input_size=None,
                 ):
        super().__init__(
            trx_encoder,
            input_size,
            hidden_size,
            type,
            bidir,
            trainable_starter,
        )
        head_params = dict(head_layers).get('CombinedTargetHeadFromRnn', None)
        self.pass_samples = head_params.get('pass_samples', True)
        self.numeric_name = list(trx_encoder.scalers.keys())[0]
        self.collect_pos, self.collect_neg = (head_params.get('pos', True), head_params.get('neg', True)) if head_params else (0, 0)
        self.eps = 1e-7

    def forward(self, x):
        amount_col = []
        for i, row in enumerate(x.payload[self.numeric_name]):
            amount_col += [list(row[:x.seq_lens[i].item()].cpu().numpy())]
        amount_col = np.array(amount_col, dtype=object)
        neg_sums = []
        pos_sums = []
        for list_row in amount_col:
            np_row = np.array(list_row)
            if self.collect_neg:
                neg_sums += [np.sum(self.transform_inv(np_row[np.where(np_row < 0)]))]
            if self.collect_pos:
                pos_sums += [np.sum(self.transform_inv(np_row[np.where(np_row >= 0)]))]
        neg_sum_logs = np.log(np.abs(np.array(neg_sums)) + self.eps)
        pos_sum_logs = np.log(np.array(pos_sums) + self.eps)

        x = super().forward(x)
        if (not self.pass_samples):
            return x
        if self.collect_neg and self.collect_pos:
            return x, neg_sum_logs, pos_sum_logs
        elif self.collect_neg:
            return x, neg_sum_logs
        elif self.collect_pos:
            return x, pos_sum_logs


from .coles_dataset import ColesDataset, ColesIterableDataset
from .coles_supervised_dataset import ColesSupervisedDataset, ColesSupervisedIterableDataset
from .coles_module import CoLESModule
from .coles_supervised_module import ColesSupervisedModule




import torch

from ptls.frames.abs_module import ABSModule
from ptls.frames.coles.losses import ContrastiveLoss
from ptls.frames.coles.metric import BatchRecallTopK
from ptls.frames.coles.sampling_strategies.hard_negative_pair_selector import HardNegativePairSelector
from ptls.nn.head import Head
from ptls.nn.seq_encoder.containers import SeqEncoderContainer


class ColesSupervisedModule(ABSModule):
    """Contrastive Learning for Event Sequences ([CoLES](https://arxiv.org/abs/2002.08232)) (unsupervised)
    with auxiliary loss based on class labels from dataset (supervised, works for labeled data)

    Subsequences are sampled from original sequence.
    Samples from the same sequence are `positive` examples
    Samples from the different sequences are `negative` examples
    Embeddings for all samples are calculated.
    Paired distances between all embeddings are calculated.
    The loss function tends to make positive distances smaller and negative ones larger.

    Parameters
        seq_encoder:
            Model which calculate embeddings for original raw transaction sequences
            `seq_encoder` is trained by `CoLESModule` to get better representations of input sequences
        head:
            Model which helps to train. Not used during inference
            Can be normalisation layer which make embedding l2 length equals 1
            Can be MLP as `projection head` like in SymCLR framework.
        loss:
            This loss applied for contrastive learning at augmentation subsample labels
            loss object from `ptls.frames.coles.losses`.
            There are paired and triplet loss. They are required sampling strategy
            from `ptls.frames.coles.sampling_strategies`. Sampling strategy takes a relevant pairs or triplets from
            pairwise distance matrix.
        l_loss:
            This loss applied for contrastive learning at auxiliary class labels
            loss object from `ptls.frames.coles.losses`.
            There are paired and triplet loss. They are required sampling strategy
            from `ptls.frames.coles.sampling_strategies`. Sampling strategy takes a relevant pairs or triplets from
            pairwise distance matrix.
        supervised_loss_w:
            weight for auxiliary losses
        validation_metric:
            Keep None. `ptls.frames.coles.metric.BatchRecallTopK` used by default.
        optimizer_partial:
            optimizer init partial. Network parameters are missed.
        lr_scheduler_partial:
            scheduler init partial. Optimizer are missed.

    """
    def __init__(self,
                 seq_encoder: SeqEncoderContainer = None,
                 head=None,
                 loss=None,
                 l_loss=None,
                 contrastive_loss_w=1.0,
                 supervised_loss_w=0.1,
                 validation_metric=None,
                 optimizer_partial=None,
                 lr_scheduler_partial=None):
        if head is None:
            head = Head(use_norm_encoder=True)

        if loss is None:
            loss = ContrastiveLoss(margin=0.5,
                                   sampling_strategy=HardNegativePairSelector(neg_count=5))

        if validation_metric is None:
            validation_metric = BatchRecallTopK(K=4, metric='cosine')

        super().__init__(validation_metric,
                         seq_encoder,
                         loss,
                         optimizer_partial,
                         lr_scheduler_partial)

        self._head = head
        self.l_loss = l_loss
        self.contrastive_loss_w = contrastive_loss_w
        self.supervised_loss_w = supervised_loss_w

    @property
    def metric_name(self):
        return 'recall_top_k'

    @property
    def is_requires_reduced_sequence(self):
        return True

    def shared_step(self, x, y, l):
        y_h = self(x)
        if self._head is not None:
            y_h = self._head(y_h)
        return y_h, y, l

    def training_step(self, batch, _):
        x = batch[0]
        y_h, y, labels = self.shared_step(*batch)

        loss = self._loss(y_h, y)  # unsupervised loss

        # supervised losses
        l_loss = 0.0
        for label_ix in range(labels.size(1)):
            l = labels[:, label_ix]
            ix_labeled = l >= 0
            if ix_labeled.sum() == 0:
                continue
            l_loss_i = self.l_loss(y_h[ix_labeled], l[ix_labeled])
            self.log(f'loss_{label_ix}/loss', l_loss_i)
            l_loss = l_loss + l_loss_i

            l_unique, l_counts = torch.unique(l, return_counts=True)
            for _l, _c in zip(l_unique, l_counts.float()):
                self.log(f'loss_{label_ix}/label_{_l.item()}_count', _c)

        self.log('loss', loss)
        self.log('seq_len', x.seq_lens.float().mean(), prog_bar=True)
        return self.contrastive_loss_w * loss + self.supervised_loss_w * l_loss
        # return self.supervised_loss_w * l_loss

    def validation_step(self, batch, _):
        y_h, y, l = self.shared_step(*batch)
        self._validation_metric(y_h, y)

    def validation_epoch_end(self, outputs):
        self.log(self.metric_name, self._validation_metric.compute(), prog_bar=True)
        self._validation_metric.reset()



# coding: utf-8
import logging
import torch

from torch.utils.data import Dataset, IterableDataset

from ptls.data_load import IterableProcessingDataset

logger = logging.getLogger(__name__)


class SplittingDataset(Dataset):
    def __init__(self, base_dataset, splitter):
        self.base_dataset = base_dataset
        self.splitter = splitter

    def __len__(self):
        return len(self.base_dataset)

    def __getitem__(self, idx):
        row = self.base_dataset[idx]

        feature_arrays = row['feature_arrays']
        local_date = row['event_time']

        indexes = self.splitter.split(local_date)
        data = [{k: v[ix] for k, v in feature_arrays.items()} for ix in indexes]
        return data


class MapSplittingDataset(Dataset):
    def __init__(self, base_dataset, splitter, a_chain):
        self.base_dataset = base_dataset
        self.splitter = splitter
        self.a_chain = a_chain

    def __len__(self):
        return len(self.base_dataset)

    def __getitem__(self, idx):
        feature_arrays = self.base_dataset[idx]
        local_date = feature_arrays['event_time']
        indexes = self.splitter.split(local_date)
        return [self.a_chain({k: v[ix] for k, v in feature_arrays.items()}) for ix in indexes]


class IterableSplittingDataset(IterableProcessingDataset):
    def __init__(self, splitter, a_chain, col_time='event_time'):
        super().__init__()

        self.splitter = splitter
        self.a_chain = a_chain
        self.col_time = col_time

    def __iter__(self):
        for row in self._src:
            local_date = row[self.col_time]
            yield [self.a_chain({k: v[ix] for k, v in row.items()}) for ix in self.splitter.split(local_date)]


class SeveralSplittingsDataset(Dataset):
    def __init__(self, base_dataset, splitters):
        self.base_dataset = base_dataset
        self.splitters = splitters

    def __len__(self):
        return len(self.base_dataset)

    def __getitem__(self, idx):
        row = self.base_dataset[idx]

        feature_arrays = row['feature_arrays']
        local_date = row['event_time']

        data = []
        for splitter in self.splitters:
            indexes = splitter.split(local_date)
            data += [{k: v[ix] for k, v in feature_arrays.items()} for ix in indexes]
        return data




import numpy as np
import torch
from collections import defaultdict
from functools import reduce

from ptls.data_load.feature_dict import FeatureDict
from ptls.data_load.padded_batch import PaddedBatch


def collate_feature_dict(batch):
    """Collate feature with arrays to padded batch

    Check feature consistency. Keys for all batch samples should be the same.
    Convert scalar value to tensors like target col

    Parameters
    ----------
    batch:
        list with feature dicts
    Returns
    -------
        PaddedBatch
    """
    new_x_ = defaultdict(list)
    for i, x in enumerate(batch):
        for k, v in x.items():
            new_x_[k].append(v)
        assert reduce(
            lambda a, b: ((a[1] is not None and a[1] == b or a[1] is None) and a[0], b),
            map(len, new_x_.values()), (True, None))[0]

    seq_col = next(k for k, v in batch[0].items() if FeatureDict.is_seq_feature(k, v))
    lengths = torch.LongTensor([len(rec[seq_col]) for rec in batch])
    new_x = {}
    for k, v in new_x_.items():
        if type(v[0]) is torch.Tensor:
            if k.startswith('target'):
                new_x[k] = torch.stack(v, dim=0)
            else:
                new_x[k] = torch.nn.utils.rnn.pad_sequence(v, batch_first=True)
        elif type(v[0]) is np.ndarray:
            new_x[k] = v  # list of arrays[object]
        else:
            v = np.array(v)
            if v.dtype.kind == 'i':
                new_x[k] = torch.from_numpy(v).long()
            elif v.dtype.kind == 'f':
                new_x[k] = torch.from_numpy(v).float()
            else:
                new_x[k] = v
    return PaddedBatch(new_x, lengths)


def collate_target(x, num=1):
    vec = np.array(x, dtype=np.float32)
    if num == 1:
        return vec.sum()
    elif abs(num) >= len(vec):
        return vec
    elif num < 0:
        return vec[:abs(num)]
    else:
        return np.hstack((vec[:num-1], vec[num-1:].sum()))[:len(vec)]



import torch
from ptls.data_load.feature_dict import FeatureDict

from ptls.data_load.utils import collate_feature_dict
from ptls.data_load.augmentations.random_slice import RandomSlice


class CpcDataset(torch.utils.data.Dataset):
    """

    Parameters
    ----------
    data:
        List with all sequences
    min_len:
        RandomSlice params.
    max_len:
        RandomSlice params.
    rate_for_min:
        RandomSlice params.
    """
    def __init__(self,
                 data,
                 min_len, max_len, rate_for_min: float = 1.0,
                 *args, **kwargs,
                 ):
        super().__init__(*args, **kwargs)

        self.data = data
        self.r_slice = RandomSlice(min_len, max_len, rate_for_min)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        feature_arrays = self.data[item]
        return self.process(feature_arrays)

    def __iter__(self):
        for feature_arrays in self.data:
            yield self.process(feature_arrays)

    def process(self, feature_arrays):
        feature_arrays = {k: v for k, v in feature_arrays.items() if FeatureDict.is_seq_feature(k, v)}
        return self.r_slice(feature_arrays)

    @staticmethod
    def collate_fn(batch):
        return collate_feature_dict(batch), None


class CpcIterableDataset(CpcDataset, torch.utils.data.IterableDataset):
    pass



import logging
import hydra
import pytorch_lightning as pl
import pyspark.sql.functions as F
import pyspark.sql.types as T
from omegaconf import DictConfig, OmegaConf
from pyspark.sql import SparkSession

logger = logging.getLogger(__name__)

class InferenceSpark(object):

    def __init__(self, spark, work_path, model_path, output_path, dataset_files,
                       col_id, pl_module_class, hidden_size, batch_size):
        self.spark           = spark
        self.work_path       = work_path
        self.model_path      = model_path
        self.output_path     = output_path
        self.dataset_files   = dataset_files
        self.col_id          = col_id
        self.pl_module_class = pl_module_class
        self.hidden_size     = hidden_size
        self.batch_size      = batch_size

    def collect_batches(self):

        df_t = self.spark.read.parquet(*[f"{self.work_path}/{i}" for i in self.dataset_files])

        columns = [i for i in df_t.columns if i not in [self.col_id, "trx_count", "target"]]

        df_t\
        .orderBy("trx_count")\
        .withColumn("group_id",F.monotonically_increasing_id())\
        .groupby(F.floor(F.col("group_id")/self.batch_size).alias("group_id"))\
        .agg(
            F.collect_list(self.col_id).alias(self.col_id),
            F.struct([F.collect_list(F.col(name)).alias(name) for name in columns]).alias("feature_arrays"),
            F.collect_list("trx_count").alias("trx_count"))\
        .write.format("parquet").mode("overwrite")\
        .save(f"{self.work_path}/data/tmp_batch_train")

        logger.info(f'broken into batches no more than {self.batch_size} and save path {self.work_path}/data/tmp_batch_train')

    def exec_inference(self):

        pl.seed_everything(42)

        model = hydra.utils.instantiate(self.pl_module_class)

        model = model.load_from_checkpoint(f"{self.work_path}/{self.model_path}")
        model.seq_encoder.is_reduce_sequence = True

        br_m = self.spark.sparkContext.broadcast(model.seq_encoder)

        def inference_func(data_feature, data_length):

            import torch
            from ptls.data_load.padded_batch import PaddedBatch
            import subprocess

            if torch.cuda.is_available():
                result = subprocess.check_output(
                    ['nvidia-smi', '--query-gpu=memory.free',
                        '--format=csv,nounits,noheader'
                    ], encoding='utf-8')
                gpu_memory = [int(x) for x in result.strip().split('\n')]
                num_cuda = gpu_memory.index(max(gpu_memory))
                device = torch.device(f"cuda:{num_cuda}")
            else:
                device = torch.device("cpu")

            data_obj = PaddedBatch(
                payload={
                    k: torch.nn.utils.rnn.pad_sequence(
                        [torch.tensor(x) for x in v], batch_first=True
                    ).to(device) for k, v in data_feature.asDict().items()
                },
                length=torch.IntTensor(data_length).to(device)
            )

            br_m.value.to(device)
            br_m.value.eval()

            with torch.no_grad():
                outputs = br_m.value(data_obj)
            torch.cuda.empty_cache()
            return outputs.cpu().numpy().tolist()

        inference_func_udf = F.udf(inference_func, T.ArrayType(T.ArrayType(T.FloatType())))

        self.spark.read.parquet(f"{self.work_path}/data/tmp_batch_train")\
        .repartition(100)\
        .select(
            F.col(self.col_id),
            inference_func_udf(
                F.col("feature_arrays"),
                F.col("trx_count")
            ).alias("inf_res")
        )\
        .write.format("parquet").mode("overwrite")\
        .save(f"{self.work_path}/data/tmp_res_emb")

        logger.info(f'distribute to cuda and make inferences and save path {self.work_path}/data/tmp_res_emb')

    def explode_embedd(self):

        self.spark.read.parquet(f"{self.work_path}/data/tmp_res_emb")\
        .withColumn("cols_zip",F.arrays_zip(self.col_id,"inf_res"))\
        .withColumn("cols_explode",F.explode("cols_zip"))\
        .select(F.col("cols_explode")[self.col_id].alias(self.col_id),
            *[F.col("cols_explode")["inf_res"][i].alias(f"v{i}") for i in range(self.hidden_size)])\
        .write.format("parquet").mode("overwrite")\
        .save(f"{self.work_path}/{self.output_path}")

        logger.info(f'explode arrays and save path {self.work_path}/{self.output_path}')


@hydra.main()
def main(conf: DictConfig):
    OmegaConf.set_struct(conf, False)
    orig_cwd = hydra.utils.get_original_cwd()

    spark = SparkSession.builder\
        .appName("spark_inference")\
        .master(f"local[{conf.inference_dataloader.loader.num_workers}]")\
        .config("spark.sql.shuffle.partitions", 100)\
        .config("spark.driver.memory", conf.spark_memory)\
        .config("spark.local.dir", f"{conf.work_path}/spark_local_dir")\
        .enableHiveSupport()\
        .getOrCreate()

    inference_obj = InferenceSpark(spark,
        conf.work_path,
        conf.model_path,
        conf.output.path,
        conf.inference_dataloader.dataset_files,
        conf.inference_dataloader.col_id,
        conf.params.pl_module_class,
        conf.params.rnn.hidden_size,
        conf.inference_dataloader.loader.batch_size
        )

    inference_obj.collect_batches()
    inference_obj.exec_inference()
    inference_obj.explode_embedd()

    spark.stop()

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)-7s %(funcName)-20s : %(message)s')
    logging.getLogger("lightning").setLevel(logging.INFO)
    main()



import numpy as np

from functools import partial
from multiprocessing import Pool
from sklearn.decomposition import PCA


def quantile_(input_array, q_count):
    return np.quantile(input_array, q=[i / q_count for i in range(1, q_count)], axis=0)


def digitize_(input_array, bins):
    return np.digitize(input_array, bins)


def embeds_quantiles(input_array: np.array, q_count: int = 0, bins: np.array = None, num_workers: int = 0):
    """Quantile-based discretization function.

    Parameters:
    -------
    input_array (np.array): Input array.
    q_count (int): Amount of quantiles. Used only if input parameter `bins` is None.
    bins (np.array):
        If None, then calculate bins as quantiles of input array,
        otherwise only apply bins to input_array. Default: None
    num_workers (int): Amount of workers in Pool in case num_workers>1, without multiprocessing otherwise. Default: 0

    Returns
    -------
    out_array (np.array of ints): discretized input_array
    bins (np.array of floats):
        Returned only if input parameter `bins` is None.
    """

    if bins is None:
        return_bins = True
        reduce_partial = partial(quantile_, q_count=q_count)
        if num_workers > 1:
            with Pool(num_workers) as p:
                bins = p.map(reduce_partial, input_array.T)
            bins = np.concatenate([x.reshape(-1, 1) for x in bins], axis=1)
        else:
            bins = quantile_(input_array, q_count)
    else:
        assert input_array.shape[1] == bins.shape[1]
        return_bins = False
        q_count = bins.shape[0]

    if num_workers > 1:
        with Pool(num_workers) as p:
            out_array = p.starmap(digitize_, zip(input_array.T, bins.T))
    else:
        out_array = map(lambda x, y: np.digitize(x, y), input_array.T, bins.T)
    out_array = np.concatenate([x.reshape(-1, 1) for x in out_array], axis=1)

    if q_count < 128:
        out_array = out_array.astype(np.int8)
    elif q_count < 256:
        out_array -= 128
        out_array = out_array.astype(np.int8)

    if return_bins:
        return out_array, bins
    else:
        return out_array


def reduce_part_(args_list):
    part, n_components, pca = args_list
    if pca is None:
        return_pca = True
        pca = PCA(n_components)
        pca.fit(part)
    else:
        return_pca = False

    transformed = pca.fit_transform(part)
    if return_pca:
        return transformed, pca
    else:
        return transformed


def pca_reduction(input_array, sub_dim: int, n_components: int = 0, pca_list=None, num_workers: int = 0):
    """PCA-based embeddings dimension reduction function.

    Split all array on parts by columns. Reduce each part with PCA. Concatenate results.

    Parameters:
    -------
    input_array (np.array): Input array.
    sub_dim (int): Dimension of each parts.
    n_components (int): Dimension of each parts after reduction. Used only if input parameter `pca_list` is not None.
    pca_list (np.array): list of PCA objects, to be applied to input_array
   num_workers (int): Amount of workers in Pool in case num_workers>1, without multiprocessing otherwise. Default: 0

    Returns
    -------
    out_array (np.array of floats): reduced input_array
    pca_list (list):
        List of PCA objects, fitted on input_array. Returned only if input parameter `pca_list` is None.
    """

    n_parts = (input_array.shape[1] - 1) // sub_dim + 1
    parts = [input_array[:, i * sub_dim: (i + 1) * sub_dim] for i in range(n_parts)]
    if pca_list is None:
        args_list = [(part, n_components, None) for part in parts]
        return_pca = True
    else:
        args_list = [(part, n_components, pca) for part, pca in zip(parts, pca_list)]
        return_pca = False

    if num_workers > 1:
        with Pool(num_workers) as p:
            transformed = p.map(reduce_part_, args_list)
    else:
        transformed = list(map(reduce_part_, args_list))
    if pca_list is None:
        transformed, pca_list = zip(*transformed)
    transformed = np.concatenate(transformed, axis=1)

    if return_pca:
        return transformed, pca_list
    else:
        return transformed



import torch

from ptls.nn import TimeStepShuffle
from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_step import SkipStepEncoder


def test_timestep_shuffle():
    t = torch.tensor([
        [[0, 0], [1, 2], [3, 4], [0, 0]],
        [[0, 0], [10, 11], [0, 0], [0, 0]],
    ])

    res = TimeStepShuffle()(PaddedBatch(t, [2, 1]))

    assert res.payload.shape == (2, 4, 2)


def test_skip_step_encoder():
    t = torch.arange(8*11*2).view(8, 11, 2)

    res = SkipStepEncoder(3)(PaddedBatch(t, [10, 9, 8, 7, 3, 2, 1, 0]))

    assert res.payload.shape == (8, 4, 2)


import torch
import warnings
from torch import nn as nn

from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from ptls.nn.seq_step import LastStepEncoder, LastMaxAvgEncoder, FirstStepEncoder
from ptls.data_load.padded_batch import PaddedBatch


# TODO: split it on GRU Rnn Encoder and LSTM Rnn Encoder
class RnnEncoder(AbsSeqEncoder):
    """Use torch recurrent layer network
    Based on `torch.nn.GRU` and `torch.nn.LSTM`

    Parameters
        input_size:
            input embedding size
        hidden_size:
            intermediate and output layer size
        type:
            'gru' or 'lstm'
            Type of rnn network
        bidir:
            Bidirectional RNN
        dropout:
            RNN dropout
        trainable_starter:
            'static' - use random learnable vector for rnn starter
            other values - use None as starter
        is_reduce_sequence:
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token

    Example:
    >>> model = RnnEncoder(
    >>>     input_size=5,
    >>>     hidden_size=6,
    >>>     is_reduce_sequence=False,
    >>> )
    >>> x = PaddedBatch(
    >>>     payload=torch.arange(4*5*8).view(4, 8, 5).float(),
    >>>     length=torch.tensor([4, 2, 6, 8])
    >>> )
    >>> out = model(x)
    >>> assert out.payload.shape == (4, 8, 6)

    """
    def __init__(self,
                 input_size=None,
                 hidden_size=None,
                 type='gru',
                 bidir=False,
                 num_layers=1,
                 dropout=0,
                 trainable_starter='static',
                 is_reduce_sequence=False,  # previous default behavior RnnEncoder
                 reducer='last_step'
                 ):
        super().__init__(is_reduce_sequence=is_reduce_sequence)

        self.hidden_size = hidden_size
        self.rnn_type = type
        self.bidirectional = bidir
        if self.bidirectional:
            warnings.warn("Backward direction in bidir RNN takes into account paddings at the end of sequences!")
        
        self.trainable_starter = trainable_starter

        # initialize RNN
        if self.rnn_type == 'lstm':
            self.rnn = nn.LSTM(
                input_size,
                self.hidden_size,
                num_layers=num_layers,
                batch_first=True,
                bidirectional=self.bidirectional,
                dropout=dropout)
        elif self.rnn_type == 'gru':
            self.rnn = nn.GRU(
                input_size,
                self.hidden_size,
                num_layers=num_layers,
                batch_first=True,
                bidirectional=self.bidirectional,
                dropout=dropout)
        else:
            raise Exception(f'wrong rnn type "{self.rnn_type}"')

        self.full_hidden_size = self.hidden_size if not self.bidirectional else self.hidden_size * 2

        # initialize starter position if needed
        if self.trainable_starter == 'static':
            num_dir = 2 if self.bidirectional else 1
            self.starter_h = nn.Parameter(torch.randn(num_dir, 1, self.hidden_size))

        if reducer == 'last_step':
            self.reducer = LastStepEncoder()
        elif reducer == 'first_step':
            self.reducer = FirstStepEncoder()
        elif reducer == 'last_max_avg':
            self.reducer = LastMaxAvgEncoder()

    def forward(self, x: PaddedBatch, h_0: torch.Tensor = None):
        """

        :param x:
        :param h_0: None or [1, B, H] float tensor
                    0.0 values in all components of hidden state of specific client means no-previous state and
                    use starter for this client
                    h_0 = None means no-previous state for all clients in batch
        :return:
        """
        shape = x.payload.size()
        assert shape[1] > 0, "Batch can'not have 0 transactions"

        # prepare initial state
        if self.trainable_starter == 'static':
            starter_h = torch.tanh(self.starter_h.expand(-1, shape[0], -1).contiguous())
            if h_0 is None:
                h_0 = starter_h
            elif h_0 is not None and not self.training:
                h_0 = torch.where(
                    (h_0.squeeze(0).abs().sum(dim=1) == 0.0).unsqueeze(0).unsqueeze(2).expand(*starter_h.size()),
                    starter_h,
                    h_0,
                )
            else:
                raise NotImplementedError('Unsupported mode: cannot mix fixed X and learning Starter')

        # pass-through rnn
        if self.rnn_type == 'lstm':
            out, _ = self.rnn(x.payload)
        elif self.rnn_type == 'gru':
            out, _ = self.rnn(x.payload, h_0)
        else:
            raise Exception(f'wrong rnn type "{self.rnn_type}"')

        out = PaddedBatch(out, x.seq_lens)
        if self.is_reduce_sequence:
            return self.reducer(out)
        return out

    @property
    def embedding_size(self):
        return self.hidden_size



from collections import OrderedDict

import torch
from torch import nn as nn

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.trx_encoder.scalers import scaler_by_name


class TrxMeanEncoder(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.scalers = OrderedDict(
            {name: scaler_by_name(scaler_name) for name, scaler_name in config['numeric_values'].items()}
        )

        self.embeddings = nn.ModuleDict()
        for name, dim in config['embeddings'].items():
            dict_len = dim['in']
            self.embeddings[name] = nn.EmbeddingBag(dict_len, dict_len, mode='mean')
            self.embeddings[name].weight = nn.Parameter(torch.diag(torch.ones(dict_len)).float())

    def forward(self, x: PaddedBatch):
        processed = []

        for field_name in self.embeddings.keys():
            processed.append(self.embeddings[field_name](x.payload[field_name]).detach())

        for value_name, scaler in self.scalers.items():
            var = scaler(x.payload[value_name].unsqueeze(-1).float())
            means = torch.tensor([e[:l].mean() for e, l in zip(var, x.seq_lens)]).unsqueeze(-1)
            processed.append(means)

        out = torch.cat(processed, -1)
        return out

    @staticmethod
    def output_size(config):
        sz = len(config['numeric_values'])
        sz += sum(dim['in'] for _, dim in config['embeddings'].items())
        return sz



import torch
import logging
import numpy as np

logger = logging.getLogger(__name__)


class FilterDataset(torch.utils.data.IterableDataset):
    def __init__(self, dataset,
                 post_processing=None,
                 shuffle_files=False,
                 shuffle_seed=42):

        self.base_dataset = dataset

        self.post_processing = post_processing
        self.shuffle_files = shuffle_files
        self.shuffle_seed = shuffle_seed
        self.rs = None

        self._worker_id = None
        self._num_workers = None
        self._shuffle_seed = None
        self._schema = None

    def _get_my_ids(self):
        return list(range(self._worker_id, len(self.base_dataset), self._num_workers))

    def _init_worker(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:  # single-process data loading, return the full iterator
            self._worker_id = 0
            self._num_workers = 1
            self._shuffle_seed = self.shuffle_seed
        else:  # in a worker process
            self._worker_id = worker_info.id
            self._num_workers = worker_info.num_workers
            self._shuffle_seed = worker_info.seed
        logger.debug(f'Started [{self._worker_id:02d}/{self._num_workers:02d}]')

    def _get_gen(self, my_ids):
        for ind in my_ids:
            rec = self.base_dataset[ind]
            rec = {k: self.to_torch(v) for k, v in rec.items()}
            yield rec

    def __iter__(self):
        self._init_worker()

        my_ids = self._get_my_ids()
        if self.shuffle_files:
            rs = np.random.RandomState(self._shuffle_seed % 2**32)
            rs.shuffle(my_ids)

        gen = self._get_gen(my_ids)
        if self.post_processing is not None:
            gen = self.post_processing(gen)
        return gen

    @staticmethod
    def to_torch(x):
        if type(x) is np.ndarray and x.dtype.kind in ('i', 'f'):
            return torch.from_numpy(x)
        return x



from functools import WRAPPER_ASSIGNMENTS
import torch

from ptls.nn.normalization import L2NormEncoder
from ptls.data_load.padded_batch import PaddedBatch


def _pb_shell(cls):
    class PBShell(cls):
        def __init__(self, *args, **kwargs):
            for attr in WRAPPER_ASSIGNMENTS:
                setattr(self, attr, getattr(cls, attr))
            super().__init__(*args, **kwargs)

        def forward(self, x: PaddedBatch):
            return PaddedBatch(super().forward(x.payload), x.seq_lens)

    return PBShell


PBDropout = _pb_shell(torch.nn.Dropout)
PBLinear = _pb_shell(torch.nn.Linear)
PBLayerNorm = _pb_shell(torch.nn.LayerNorm)
PBReLU = _pb_shell(torch.nn.ReLU)
PBL2Norm = _pb_shell(L2NormEncoder)



from .trx_encoder import (
    TrxEncoder, TabFormerFeatureEncoder, TrxEncoderOhe,
)

from .seq_encoder import (
    RnnEncoder,  TransformerEncoder, LongformerEncoder,
    RnnSeqEncoder, TransformerSeqEncoder, LongformerSeqEncoder, AggFeatureSeqEncoder,
    GptEncoder
)

from .pb import PBDropout, PBLinear, PBL2Norm, PBLayerNorm, PBReLU

from .pb_feature_extract import PBFeatureExtract

from .head import Head

from .normalization import L2NormEncoder

from .binarization import BinarizationLayer

from .seq_step import FirstStepEncoder, LastStepEncoder, TimeStepShuffle, SkipStepEncoder



import random
import torch
from transformers import GPT2Config, GPT2Model

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from ptls.nn.seq_step import LastStepEncoder

class GptEncoder(AbsSeqEncoder):
    """Used huggingface implementation of GPT decoder
    Based on `transformers.GPT2Model`

    Parameters
    ----------
        n_embd:
            input embedding size.
            Equals intermediate and output layer size cause transformer don't change vector dimentions
        n_head:
            The number of heads in the multiheadattention models
        n_inner:
            The dimension of the feedforward network model
        n_layer:
            The number of sub-decoder-layers in the decoder
        activation_function:
            Activation function is used.
        n_positions:
            The possible maximum sequence length for positional encoding
        resid_pdrop:
            Dropout probability of residual connections of decoder layers.
        embd_pdrop:
            Dropout probability of embeddings.
        attn_pdrop:
            Dropout probability of attention matrix.
        use_positional_encoding (bool):
            Use or not positional encoding
        use_start_random_shift (bool):
            True - starting pos of positional encoding randomly shifted when training
            This allow to train transformer with all range of positional encoding values
            False - starting pos is not shifted.
        is_reduce_sequence (bool):
            False - returns PaddedBatch with all transactions embeddings
            True - returns last embedding of sequence 
    """

    def __init__(self,
                 n_embd=768,
                 n_layer=12,
                 n_head=12,
                 n_inner=2048,
                 activation_function="gelu_new",
                 resid_pdrop=0.1,
                 embd_pdrop=0.1,
                 attn_pdrop=0.1,
                 n_positions=1024,
                 use_positional_encoding=True,
                 use_start_random_shift=True,
                 is_reduce_sequence=False,
                 ):
        super().__init__(is_reduce_sequence=is_reduce_sequence)

        self.n_embd = n_embd
        self.n_positions = n_positions
        self.use_positional_encoding = use_positional_encoding
        self.use_start_random_shift = use_start_random_shift
        
        if is_reduce_sequence:
            self.last_step = LastStepEncoder()
        
        self.transf = GPT2Model(
            config=GPT2Config(
                n_embd=n_embd,
                n_layer=n_layer,
                n_head=n_head,
                n_inner=n_inner,
                activation_function=activation_function,
                resid_pdrop=resid_pdrop,
                embd_pdrop=embd_pdrop,
                attn_pdrop=attn_pdrop,
                n_positions=n_positions,
                vocab_size=4,
            ),
        )
        
    def forward(self, x: PaddedBatch):
        B, T, H = x.payload.size()
        device = x.device

        if self.training and self.use_start_random_shift:
            start_pos = random.randint(0, self.n_positions - T - 1)
        else:
            start_pos = 0

        inputs_embeds = x.payload
        attention_mask = x.seq_len_mask.float()

        if self.use_positional_encoding:
            position_ids = torch.arange(T, device=device, dtype=torch.long).view(1, -1).expand(B, T) + start_pos
        else:
            position_ids = torch.zeros(B, T, device=device, dtype=torch.long)

        out = self.transf(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
        ).last_hidden_state

        out = PaddedBatch(out, x.seq_lens)

        if self.is_reduce_sequence:
            return self.last_step(out)
        return out

    @property
    def embedding_size(self):
        return self.n_embd



import pytest
import torch

from ptls.data_load import PaddedBatch
from ptls.nn.trx_encoder.scalers import IdentityScaler
from ptls.nn.trx_encoder.trx_encoder_base import TrxEncoderBase


def test_get_category_indexes_clip():
    m = TrxEncoderBase(
        embeddings={'mcc': {'in': 4, 'out': 5}},
    )
    x = PaddedBatch({
            'mcc': torch.arange(6).unsqueeze(0),
        },
        torch.tensor([6]),
    )
    out = m.get_category_indexes(x, 'mcc')
    exp = torch.LongTensor([[0, 1, 2, 3, 3, 3]])
    torch.testing.assert_close(out, exp)


def test_get_category_indexes_assert():
    m = TrxEncoderBase(
        embeddings={'mcc': {'in': 4, 'out': 5}},
        out_of_index='assert',
    )
    x = PaddedBatch({
            'mcc': torch.arange(6).unsqueeze(0),
        },
        torch.tensor([6]),
    )
    with pytest.raises(IndexError):
        _ = m.get_category_indexes(x, 'mcc')


def test_get_category_embeddings():
    m = TrxEncoderBase(
        embeddings={'mcc': {'in': 4, 'out': 5}},
    )
    x = PaddedBatch({
            'mcc': torch.arange(6).unsqueeze(0),
        },
        torch.tensor([6]),
    )
    out = m.get_category_embeddings(x, 'mcc')
    assert out.size() == (1, 6, 5)
    torch.testing.assert_close(out[0, 3], out[0, 5])


def test_get_numeric_scaled():
    m = TrxEncoderBase(
        numeric_values={'amount': 'log'},
    )
    x = PaddedBatch({
            'amount': torch.arange(6).unsqueeze(0),
        },
        torch.tensor([6]),
    )
    out = m.get_numeric_scaled(x, 'amount')
    torch.testing.assert_close(out.expm1(), torch.arange(6).float().view(1, 6, 1))


def test_get_numeric_scaled_with_alias():
    m = TrxEncoderBase(
        numeric_values={'amount_x': IdentityScaler(col_name='amount')},
    )
    x = PaddedBatch({
            'amount': torch.arange(6).unsqueeze(0),
        },
        torch.tensor([6]),
    )
    out = m.get_numeric_scaled(x, 'amount_x')
    torch.testing.assert_close(out, torch.arange(6).float().view(1, 6, 1))


def test_output_size():
    m = TrxEncoderBase(
        embeddings={'mcc': {'in': 4, 'out': 5}},
        numeric_values={'amount': 'log'},
    )
    assert m.numerical_size == 1
    assert m.embedding_size == 5
    assert m.output_size == 6


def test_disable_embedding1():
    m = TrxEncoderBase(
        embeddings={
            'mcc': {'in': 4, 'out': 5},
            'code': {'in': 0}
        },
    )
    assert m.embedding_size == 5


def test_disable_embedding2():
    m = TrxEncoderBase(
        embeddings={
            'mcc': {'in': 4, 'out': 5},
            'code': {'in': 8, 'out': 0},
        },
    )
    assert m.embedding_size == 5


def test_disable_embedding3():
    m = TrxEncoderBase(
        embeddings={
            'mcc': {'in': 4, 'out': 5},
            'code': {'disabled': True},
        },
    )
    assert m.embedding_size == 5


def test_disable_numerical():
    m = TrxEncoderBase(
        numeric_values={
            'amount': 'log',
            'amount_log': 'none',
        },
    )
    assert m.numerical_size == 1


def test_output_size_extended():
    m = TrxEncoderBase(
        embeddings={
            'mcc': {'in': 4, 'out': 5},
            'currency': torch.nn.Embedding(5, 3),
        },
        numeric_values={
            'amount': 'log',
            'amount_log': IdentityScaler(col_name='amount'),
        },
    )
    assert m.numerical_size == 2
    assert m.embedding_size == 8
    assert m.output_size == 10



import logging

from torch.utils.data import Dataset

logger = logging.getLogger(__name__)


class TargetEnumeratorDataset(Dataset):
    def __init__(self, base_dataset):
        self.base_dataset = base_dataset

    def __len__(self):
        return len(self.base_dataset)

    def __getitem__(self, idx):
        row = self.base_dataset[idx]
        data = [(x, idx) for x in row]
        return data



from itertools import chain

import pyspark
import pyspark.sql.functions as F
from pyspark.sql import Window

from ptls.preprocessing.base.col_category_transformer import ColCategoryTransformer
from ptls.preprocessing.pyspark.col_transformer import ColTransformerPysparkMixin


class FrequencyEncoder(ColTransformerPysparkMixin, ColCategoryTransformer):
    """Use frequency encoding for categorical field

    Let's `col_name_original` value_counts looks like this:
    cat value: records counts in dataset
          aaa:  100
          bbb:  50
          nan:  10
          ccc:  1

    Mapping will use this order to enumerate embedding indexes for category values:
    cat value: embedding id
    <padding token>: 0
                aaa: 1
                bbb: 2
                nan: 3
                ccc: 4
     <other values>: 5

    `dictionary_size` will be 6

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Target column name. Transformed column will be placed here
        If `col_name_target is None` then original column will be replaced by transformed values.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.
    max_cat_num:
        Maximum category number
    """
    def __init__(self,
                 col_name_original: str,
                 col_name_target: str = None,
                 is_drop_original_col: bool = True,
                 max_cat_num: int = 10000,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target=col_name_target,
            is_drop_original_col=is_drop_original_col,
        )

        self.mapping = None
        self.other_values_code = None
        self.max_cat_num = max_cat_num

    def get_col(self, x: pyspark.sql.DataFrame):
        return x.withColumn(self.col_name_target,
                          F.coalesce(F.col(self.col_name_original).cast('string'), F.lit('#EMPTY')))

    def fit(self, x: pyspark.sql.DataFrame):
        super().fit(x)

        df = self.get_col(x)
        df_encoder = df.groupby(self.col_name_target).agg(F.count(F.lit(1)).alias('_cnt'))
        df_encoder = df_encoder.withColumn('_rn',
                                           F.row_number().over(Window.partitionBy().orderBy(F.col('_cnt').desc())))
        df_encoder = df_encoder.filter(F.col('_rn') <= self.max_cat_num)

        self.mapping = {row[self.col_name_target]: row['_rn'] for row in df_encoder.collect()}
        self.other_values_code = len(self.mapping) + 1
        return self

    @property
    def dictionary_size(self):
        return self.other_values_code + 1

    def transform(self, x: pyspark.sql.DataFrame):
        df = self.get_col(x)

        mapping_expr = F.create_map([F.lit(x) for x in chain(*self.mapping.items())])
        df = df.withColumn(self.col_name_target, mapping_expr[F.col(self.col_name_target)])
        df = df.fillna(value=self.other_values_code, subset=[self.col_name_target])

        x = super().transform(df)
        return x



import torch
from torch import nn as nn


class NoisyEmbedding(nn.Embedding):
    """
    Embeddings with additive gaussian noise with mean=0 and user-defined variance.
    *args and **kwargs defined by usual Embeddings
    Args:
        noise_scale (float): when > 0 applies additive noise to embeddings.
            When = 0, forward is equivalent to usual embeddings.
        dropout (float): probability of embedding axis to be dropped. 0 means no dropout at all.
        spatial_dropout (bool): whether to dropout full dimension of embedding in the whole sequence.

    For other parameters defenition look at nn.Embedding help
    """

    def __init__(self, num_embeddings, embedding_dim, padding_idx=None, max_norm=None,
                 norm_type=2.0, scale_grad_by_freq=False, sparse=False, _weight=None,
                 noise_scale=0, dropout=0, spatial_dropout=False):
        if max_norm is not None:
            raise AttributeError("Please don't use embedding normalisation. "
                                 "https://github.com/pytorch/pytorch/issues/44792")

        super().__init__(num_embeddings, embedding_dim, padding_idx, max_norm,
                         norm_type, scale_grad_by_freq, sparse, _weight)
        self.noise = torch.distributions.Normal(0, noise_scale) if noise_scale > 0 else None
        self.scale = noise_scale
        self.spatial_dropout = spatial_dropout
        self.dropout = nn.Dropout2d(dropout) if spatial_dropout else nn.Dropout(dropout)

        # This is workaround for https://github.com/pytorch/pytorch/issues/44792
        # Weight are normalized after forward pass
        _ = super().forward(torch.arange(num_embeddings))

    def forward(self, x):
        x = super().forward(x)
        if self.spatial_dropout:
            x = self.dropout(x.permute(0, 2, 1).unsqueeze(3)).squeeze(3).permute(0, 2, 1)
        else:
            x = self.dropout(x)
        if self.training and self.scale > 0:
            x += self.noise.sample((self.weight.shape[1], )).to(self.weight.device)
        return x



"""These test are inspired by `docs/nn/head.md`
"""

import torch

from ptls.nn import Head


def test_empty_layer():
    x = torch.randn(10, 8)
    model = Head()
    y = model(x)
    torch.testing.assert_allclose(x, y)


def test_l2_norm():
    x = torch.randn(10, 8)
    model = Head(use_norm_encoder=True)
    y = model(x)
    assert y.size() == (10, 8)


def test_binary_classification():
    x = torch.randn(10, 8)
    model = Head(objective='classification', input_size=8)
    y = model(x)
    assert y.size() == (10,)


def test_multiclass_classification():
    x = torch.randn(10, 8)
    model = Head(objective='classification', input_size=8, num_classes=3)
    y = model(x)
    assert y.size() == (10, 3)


def test_mlp_binary_classification():
    x = torch.randn(10, 8)
    model = Head(objective='classification', input_size=8, hidden_layers_sizes=[12, 36])
    y = model(x)
    assert y.size() == (10,)


def test_regression():
    x = torch.randn(10, 8)
    model = Head(objective='regression', input_size=8)
    y = model(x)
    assert y.size() == (10,)
    model = Head(objective='regression', input_size=8, num_classes=3)
    y = model(x)
    assert y.size() == (10, 3)


def test_softplus():
    x = torch.randn(10, 8)
    model = Head(objective='softplus', input_size=8)
    y = model(x)
    assert y.size() == (10,)
    model = Head(objective='softplus', input_size=8, num_classes=3)
    y = model(x)
    assert y.size() == (10, 3)



import argparse
import datetime
import logging
import os
import pickle
from random import Random

import numpy as np
import pandas as pd

from ptls.preprocessing.util import pd_hist

logger = logging.getLogger(__name__)


def parse_args(args=None):
    parser = argparse.ArgumentParser()

    parser.add_argument('--data_path', type=os.path.abspath)
    parser.add_argument('--trx_files', nargs='+')
    parser.add_argument('--target_files', nargs='*', default=[])

    parser.add_argument('--print_dataset_info', action='store_true', default=True)
    parser.add_argument('--col_client_id', type=str)
    parser.add_argument('--cols_event_time', nargs='+')
    parser.add_argument('--cols_category', nargs='*', default=[])
    parser.add_argument('--cols_log_norm', nargs='*', default=[])
    parser.add_argument('--col_target', required=False, type=str)
    parser.add_argument('--test_size', type=float, default=0.1)
    parser.add_argument('--salt', type=int, default=42)

    parser.add_argument('--output_train_path', type=os.path.abspath)
    parser.add_argument('--output_test_path', type=os.path.abspath)
    parser.add_argument('--output_test_ids_path', type=os.path.abspath)
    parser.add_argument('--log_file', type=os.path.abspath)

    args = parser.parse_args(args)
    logger.info('Parsed args:\n' + '\n'.join([f'  {k:15}: {v}' for k, v in vars(args).items()]))
    return args


def load_source_data(data_path, trx_files):
    data = []
    for file in trx_files:
        file_path = os.path.join(data_path, file)
        df = pd.read_csv(file_path)
        data.append(df)
        logger.info(f'Loaded {len(df)} rows from "{file_path}"')

    data = pd.concat(data, axis=0)
    logger.info(f'Loaded {len(data)} rows in total')
    return data


def encode_col(col):
    col = col.astype(str)
    return col.map({k: i + 1 for i, k in enumerate(col.value_counts().index)})


def trx_to_features(df_data, print_dataset_info,
                    col_client_id, cols_event_time, cols_category, cols_log_norm):
    def copy_time(rec):
        rec['event_time'] = rec['feature_arrays']['event_time']
        del rec['feature_arrays']['event_time']
        return rec

    def _td_default(df, cols_event_time):
        df_event_time = df[cols_event_time].drop_duplicates()
        df_event_time = df_event_time.sort_values(cols_event_time)
        df_event_time['event_time'] = np.arange(len(df_event_time))
        df = pd.merge(df, df_event_time, on=cols_event_time)
        logger.info('Default time transformation')
        return df

    def _td_float(df, col_event_time):
        df['event_time'] = df[col_event_time].astype(float)
        logger.info('To-float time transformation')
        return df

    def _td_gender(df, col_event_time):
        """Gender-dataset-like transformation

        'd hh:mm:ss' -> float where integer part is day number and fractional part is seconds from day begin
        '1 00:00:00' -> 1.0
        '1 12:00:00' -> 1.5
        '1 01:00:00' -> 1 + 1 / 24
        '2 23:59:59' -> 1.99
        '432 12:00:00' -> 432.5

        :param df:
        :param col_event_time:
        :return:
        """
        padded_time = df[col_event_time].str.pad(15, 'left', '0')
        day_part = padded_time.str[:6].astype(float)
        time_part = pd.to_datetime(padded_time.str[7:], format='%H:%M:%S').values.astype(int) // 1e9
        time_part = time_part % (24 * 60 * 60) / (24 * 60 * 60)
        df['event_time'] = day_part + time_part
        logger.info('Gender-dataset-like time transformation')
        return df

    if print_dataset_info:
        logger.info(f'Found {df_data[col_client_id].nunique()} unique clients')

    # event_time mapping
    if cols_event_time[0][0] == '#':
        if cols_event_time[0] == '#float':
            df_data = _td_float(df_data, cols_event_time[1])
        elif cols_event_time[0] == '#gender':
            df_data = _td_gender(df_data, cols_event_time[1])
        else:
            raise NotImplementedError(f'Unknown type of data transformation: "{cols_event_time[0]}"')
    else:
        df_data = _td_default(df_data, cols_event_time)

    for col in cols_category:
        df_data[col] = encode_col(df_data[col])
        if print_dataset_info:
            logger.info(f'Encoder stat for "{col}":\ncodes | trx_count\n{pd_hist(df_data[col], col)}')

    for col in cols_log_norm:
        df_data[col] = np.log1p(abs(df_data[col])) * np.sign(df_data[col])
        df_data[col] /= abs(df_data[col]).max()
        if print_dataset_info:
            logger.info(f'Encoder stat for "{col}":\ncodes | trx_count\n{pd_hist(df_data[col], col)}')

    if print_dataset_info:
        df = df_data.groupby(col_client_id)['event_time'].count()
        logger.info(f'Trx count per clients:\nlen(trx_list) | client_count\n{pd_hist(df, "trx_count")}')

    # column filter
    used_columns = [col for col in df_data.columns
                    if col in cols_category + cols_log_norm + ['event_time', col_client_id]]

    logger.info('Feature collection in progress ...')
    features = df_data[used_columns] \
        .assign(et_index=lambda x: x['event_time']) \
        .set_index([col_client_id, 'et_index']).sort_index() \
        .groupby(col_client_id).apply(lambda x: {k: np.array(v) for k, v in x.to_dict(orient='list').items()}) \
        .rename('feature_arrays').reset_index().to_dict(orient='records')

    features = [copy_time(r) for r in features]

    if print_dataset_info:
        feature_names = list(features[0]['feature_arrays'].keys())
        logger.info(f'Feature names: {feature_names}')

    logger.info(f'Prepared features for {len(features)} clients')
    return features


def update_with_target(features, data_path, target_files, col_client_id, col_target):
    df_target = pd.concat([pd.read_csv(os.path.join(data_path, file)) for file in target_files])
    df_target = df_target.set_index(col_client_id)
    d_clients = df_target.to_dict(orient='index')
    logger.info(f'Target loaded for {len(d_clients)} clients')

    features = [
        dict([('target', d_clients.get(rec[col_client_id], {}).get(col_target))] + list(rec.items()))
        for rec in features
    ]
    logger.info(f'Target updated for {len(features)} clients')
    return features


def split_dataset(all_data, test_size, data_path, target_files, col_client_id, salt):
    df_target = pd.concat([pd.read_csv(os.path.join(data_path, file)) for file in target_files])
    s_clients = set(df_target[col_client_id].tolist())

    # shuffle client list
    s_all_data_clients = set(rec[col_client_id] for rec in all_data)
    s_clients = (cl_id for cl_id in s_clients if cl_id in s_all_data_clients)
    s_clients = sorted(s_clients)
    s_clients = [cl_id for cl_id in s_clients]
    Random(salt).shuffle(s_clients)

    # split client list
    Nrows_test = int(len(s_clients) * test_size)
    s_clients_train = s_clients[:-Nrows_test]
    s_clients_test = s_clients[-Nrows_test:]

    # split data
    labeled_train = [rec for rec in all_data if rec[col_client_id] in s_clients_train]
    labeled_test = [rec for rec in all_data if rec[col_client_id] in s_clients_test]
    unlabeled = [rec for rec in all_data if rec[col_client_id] not in s_clients]
    train = labeled_train + unlabeled
    test = labeled_test

    logger.info(f'Train size: {len(train)} clients')
    logger.info(f'Test size: {len(test)} clients')

    return train, test


def save_features(df_data, save_path):
    with open(save_path, 'wb') as f:
        pickle.dump(df_data, f)
    logger.info(f'Saved to: "{save_path}"')


if __name__ == '__main__':
    _start = datetime.datetime.now()
    config = parse_args()

    if config.log_file is not None:
        handlers = [logging.StreamHandler(), logging.FileHandler(config.log_file, mode='w')]
    else:
        handlers = None
    logging.basicConfig(level=logging.INFO, format='%(funcName)-20s   : %(message)s',
                        handlers=handlers)

    source_data = load_source_data(
        data_path=config.data_path,
        trx_files=config.trx_files,
    )

    client_features = trx_to_features(
        df_data=source_data,
        print_dataset_info=config.print_dataset_info,
        col_client_id=config.col_client_id,
        cols_event_time=config.cols_event_time,
        cols_category=config.cols_category,
        cols_log_norm=config.cols_log_norm,
    )

    if len(config.target_files) > 0 and config.col_target is not None:
        client_features = update_with_target(
            features=client_features,
            data_path=config.data_path,
            target_files=config.target_files,
            col_client_id=config.col_client_id,
            col_target=config.col_target,
        )

    if config.test_size > 0:
        train, test = split_dataset(
            all_data=client_features,
            test_size=config.test_size,
            data_path=config.data_path,
            target_files=config.target_files,
            col_client_id=config.col_client_id,
            salt=config.salt,
        )
    else:
        train = client_features

    save_features(
        df_data=train,
        save_path=config.output_train_path,
    )

    if config.test_size > 0:
        save_features(
            df_data=test,
            save_path=config.output_test_path,
        )
        test_ids = pd.DataFrame({config.col_client_id: [rec[config.col_client_id] for rec in test]})
        test_ids.to_csv(config.output_test_ids_path, index=False)

    _duration = datetime.datetime.now() - _start
    logger.info(f'Data collected in {_duration.seconds} sec ({_duration})')
